"""Security Management Protocol

BleSMP provides these different pairing strategies:

- "Legacy Just Works"
- "Legacy Passkey Entry"
- "LESC JustWorks"
- "LESC Numeric Comparison"
- "LESC Passkey Entry"

"""
# TODO
# refactoring, maybe some states can be merged

from struct import pack, unpack
from random import randint
from time import sleep
from typing import Optional

from scapy.layers.bluetooth import SM_Pairing_Request, SM_Pairing_Response, SM_Hdr,\
    SM_Confirm, SM_Random, SM_Failed, SM_Encryption_Information, SM_Master_Identification, \
    SM_Identity_Information, SM_Signing_Information, SM_Identity_Address_Information, \
    SM_DHKey_Check, SM_Public_Key
from whad.scapy.layers.bluetooth import SM_Security_Request
from whad.ble.crypto import LinkLayerCryptoManager, generate_random_value, c1, s1, f4, g2, \
    f5, f6, generate_public_key_from_coordinates, generate_diffie_hellman_shared_secret, \
    generate_p256_keypair
from whad.hub.ble.bdaddr import BDAddress
from whad.ble.stack.smp.constants import *
from whad.ble.stack.smp.parameters import Pairing, PairingCustomFunctions
from whad.ble.stack.smp.exceptions import SMInvalidParameterFormat, SMInvalidCustomFunction

from whad.common.stack import Layer, alias, source, instance, LayerState, state

import json
import logging
logger = logging.getLogger(__name__)

class BLEKey:
    def __init__(self, key, key_size=None, type="LTK"):
        self.__key = key
        self.__key_size = 8*len(key) if key_size is None else key_size
        self.__type = type

    @property
    def value(self):
        return self.__key

    @property
    def key_size(self):
        return self.__key_size

    def __eq__(self, other):
        if isinstance(other, bytes):
            return other == self.value
        else:
            return (
                other.value == self.value
            )

    def __repr__(self):
        return "%s('%s' - %d bits)" % (self.__type, self.__key.hex().lower(), self.__key_size)

class LongTermKey(BLEKey):
    def __init__(self, key, rand=b"\x00"*8, ediv=0, key_size=None):
        super().__init__(key, key_size=key_size, type="LTK")
        self.__rand = rand
        self.__ediv = ediv

    @property
    def rand(self):
        return self.__rand

    @property
    def ediv(self):
        return self.__ediv

    def __eq__(self, other):
        if isinstance(other, bytes):
            return other == self.value
        else:
            return (
                other.rand == self.rand and
                other.ediv == self.ediv and
                other.value == self.value
            )

    def __repr__(self):
        return "LTK('%s' - %d bits / rand = %s / ediv = %s)" % (self.value.hex().lower(), self.key_size, self.rand.hex(), hex(self.ediv))

class IdentityResolvingKey(BLEKey):
    def __init__(self, key, key_size=None):
        super().__init__(key, key_size=key_size, type="IRK")


class ConnectionSignatureResolvingKey(BLEKey):
    def __init__(self, key, key_size=None):
        super().__init__(key, key_size=key_size, type="CSRK")


class CryptographicMaterial:

    @classmethod
    def from_json(cls, json_value):
        report = json.loads(json_value)

        cm = CryptographicMaterial(

            BDAddress(report["address"], random = (report["address_type"] == 1)),

            ltk = LongTermKey(
                bytes.fromhex(report["ltk"]),
                rand=bytes.fromhex(report["rand"]),
                ediv=report["ediv"]
            ) if "ltk" in report else None,

            irk = IdentityResolvingKey(
                bytes.fromhex(report["irk"])
            ) if "irk" in report else None,

            csrk = ConnectionSignatureResolvingKey(
                bytes.fromhex(report["csrk"])
            ) if "csrk" in report else None
        )
        return cm

    def to_json(self):
        report = {
            "address" : ":".join(["{:02x}".format(i) for i in self.__address.value[::-1]]),
            "address_type" : self.__address.type
        }
        if self.has_ltk():
            report["ltk"] = self.ltk.value.hex()
            report["rand"] = self.ltk.rand.hex()
            report["ediv"] = self.ltk.ediv
        if self.has_irk():
            report["irk"] = self.irk.value.hex()
        if self.has_csrk():
            report["csrk"] = self.csrk.value.hex()

        return json.dumps(report)


    def __init__(self, address: BDAddress, authenticated: bool = False, ltk: Optional[LongTermKey] = None, irk=None, csrk=None):
        self.__address = address
        self.__authenticated = authenticated

        if isinstance(ltk, bytes):
            self.__ltk = LongTermKey(ltk)
        else:
            self.__ltk = ltk

        if isinstance(irk, bytes):
            self.__irk = IdentityResolvingKey(irk)
        else:
            self.__irk = irk

        if isinstance(csrk, bytes):
            self.__csrk = ConnectionSignatureResolvingKey(csrk)
        else:
            self.__csrk = csrk

    def is_authenticated(self):
        return self.__authenticated

    def has_ltk(self):
        return self.__ltk is not None

    def has_irk(self):
        return self.__irk is not None

    def has_csrk(self):
        return self.__csrk is not None

    @property
    def ltk(self):
        return self.__ltk

    @property
    def irk(self):
        return self.__irk

    @property
    def csrk(self):
        return self.__csrk

    @property
    def address(self):
        return self.__address


    def __repr__(self):
        output = (
            "CryptographicMaterial(" +
            ", ".join(
                ["address = " + str(self.__address)] +
                ["authenticated = " + ("yes" if self.is_authenticated() else "no")] +
                ([("ltk = " + str(self.ltk))] if self.ltk is not None else []) +
                ([("irk = " + str(self.irk))] if self.irk is not None else []) +
                ([("csrk = " + str(self.csrk))] if self.csrk is not None else [])

            ) + ")"
        )
        return output

class CryptographicDatabase:

    def to_json(self):
        entries = []
        for entry in self.__entries:
            entries += [json.loads(entry.to_json())]
        return json.dumps(entries)

    def __init__(self, *entries, from_json=None):
        self.__entries = []
        if from_json is not None:
            json_list = json.loads(from_json)
            for json_entry in json_list:
                self.__entries.append(
                    CryptographicMaterial.from_json(
                        json.dumps(json_entry)
                    )
                )
        else:
            if len(entries) > 0:
                for entry in entries:
                    self.__entries.append(entry)

    def get(self, **kwargs):
        filter = {}
        for name, value in kwargs.items():
            filter[name] = value

        for entry in self.__entries:
            if all(
                [
                    hasattr(entry, filter_name) and getattr(entry, filter_name) == filter_value
                    for filter_name, filter_value in filter.items()
                ]):
                return entry

        return None

    def __repr__(self):
        output = (
                    "CryptographicDatabase(" +
                    ("\n" if len(self.__entries) > 0 else "") +
                    ",\n".join([str(e) for e in self.__entries]) +
                    ("\n" if len(self.__entries) > 0 else "") +
                    ")"
        )
        return output

    def remove(self, **kwargs):
        entry = self.get(**kwargs)
        if entry is None:
            return False
        self.__entries.remove(entry)

    def add(self, address, authenticated=False, ltk=None , rand=None, ediv=None, irk=None, csrk=None):

        long_term_key = None
        identity_resolving_key = None
        connection_signature_resolving_key = None

        if ltk is not None:
            if rand is None or ediv is None:
                long_term_key = LongTermKey(ltk)
            else:
                long_term_key = LongTermKey(ltk, rand=rand, ediv=ediv)
        if irk is not None:
            identity_resolving_key = IdentityResolvingKey(irk)
        if csrk is not None:
            connection_signature_resolving_key = ConnectionSignatureResolvingKey(csrk)

        bd_address = None
        if isinstance(address, BDAddress):
            bd_address = address
        elif isinstance(address, bytes):
            bd_address = BDAddress.from_bytes(address)
        elif isinstance(address, str):
            bd_address = BDAddress(address)
        else:
            return False

        cm = CryptographicMaterial(
            bd_address,
            authenticated=authenticated,
            ltk=long_term_key,
            irk=identity_resolving_key,
            csrk=connection_signature_resolving_key
        )
        self.__entries.append(cm)
        return True

class SM_Peer(object):
    """
    SM_Peer stores all pairing-specific data regarding a peer.
    """


    def __init__(self, address: BDAddress):
        """Instantiate a SM_Peer.

        By default, peers are configured to only accept Legacy JustWorks pairing.
        """

        logger.info("Initiate SM_Peer object for peer %s", address)

        # Peer address and address type
        self.__address = address

        # Key distribution (by default, corresponds to Legacy JustWorks)
        self.__kd_link_key = True
        self.__kd_sign_key = True
        self.__kd_enc_key = True
        self.__kd_id_key = True

        # Default security parameters
        self.__oob = False
        self.__bonding = False
        self.__mitm = False
        self.__lesc = True
        self.__keypress = False
        self.__ct2 = False
        self.__max_key_size = 16

        # Crypto parameters
        self.__confirm = None
        self.__rand = None
        self.__stk = None
        self.__ltk = None

        # Distribution parameters
        self.__dist_ltk = False
        self.__dist_ediv_rand = False
        self.__dist_irk = False
        self.__dist_csrk = False

        self.__distributed_ltk = None
        self.__distributed_rand = None
        self.__distributed_ediv = None
        self.__distributed_address = None
        self.__distributed_address_type = None
        self.__distributed_irk  = None
        self.__distributed_csrk = None

        # Pairing method
        self.__pairing_method = PM_LEGACY_JUSTWORKS

        # IO Capabilities
        self.__io_cap = IOCAP_DISPLAY_ONLY

    @property
    def address(self):
        return self.__address.value

    @property
    def address_type(self):
        return self.__address.type

    @property
    def bd_address(self):
        return self.__address

    def support_lesc(self):
        return self.__lesc

    def support_oob(self):
        return self.__oob

    def requires_bonding(self):
        return self.__bonding

    def support_mitm(self):
        return self.__mitm

    def set_security_parameters(self,
        lesc=None, mitm=None, bonding=None, keypress=None,
        ct2=None, max_key_size=None, oob=None):
        """Set security parameters for the current SM peer.

        :param lesc: Enable/disable LE Secure Connection for this peer
        :type lesc: bool, optional
        :param mitm: Enable/disable MITM for this peer
        :type mitm: bool, optional
        :param bonding: Enable/disable bonding for this peer
        :type bonding: bool, optional
        :param keypress: Enable/disable keypress notification for this peer
        :type keypress: bool, optional
        :param ct2: Enable/disable ct2 (h7 function support) for this peer
        :type ct2: bool, optional
        :param max_key_size: Set max key size for this peer
        :type ct2: int, optional
        :param oob: Use OOB data
        :type oob: bool, optional
        """
        if isinstance(lesc, bool):
            self.__lesc = lesc
        if isinstance(mitm, bool):
            self.__mitm = mitm
        if isinstance(bonding, bool):
            self.__bonding = bonding
        if isinstance(keypress, bool):
            self.__keypress = keypress
        if isinstance(ct2, bool):
            self.__ct2 = ct2
        if isinstance(oob, bool):
            self.__oob = oob
        if max_key_size is not None and isinstance(max_key_size, int):
            self.__max_key_size = max_key_size

    @property
    def iocap(self):
        """Return this peer IO capabilities
        """
        return self.__io_cap

    @iocap.setter
    def iocap(self, value):
        if value in [
            IOCAP_DISPLAY_ONLY,
            IOCAP_DISPLAY_YESNO,
            IOCAP_KEYBD_DISPLAY,
            IOCAP_KEYBD_ONLY,
            IOCAP_NOINPUT_NOOUTPUT
            ]:

            self.__io_cap = value
        else:
            raise SMInvalidParameterFormat


    @property
    def oob(self):
        return OOB_DISABLED if not self.__oob else OOB_ENABLED

    @property
    def authentication(self):
        """Rebuild authentication byte from internal status

        :return: Pairing Request/Response AuthReq value
        :rtype: int
        """
        flags = 0x00
        if self.__bonding:
            flags |= 0x01
        if self.__mitm:
            flags |= 0x04
        if self.__lesc:
            flags |= 0x08
        if self.__keypress:
            flags |= 0x10
        if self.__ct2:
            flags |= 0x20
        return flags

    ##
    # Key distribution
    ##

    def distribute_keys(self, link_key=False, sign_key=False, id_key=False, enc_key=False):
        """Mark keys to be distributed to peer.

        :param bool ltk: Will distribute LTK if set to True
        :param bool ediv: Will distribute EDIV if set to True
        :param bool rand: Will distribute RAND if set to True
        :param bool irk: Will distribute IRK if set to True
        :param bool address: Will distribute address if set to True
        :param bool csrk: Will distribute CSRK if set to True
        """
        self.__kd_link_key = link_key
        self.__kd_sign_key = sign_key
        self.__kd_id_key = id_key
        self.__kd_enc_key = enc_key

        if self.__pairing_method in [PM_LEGACY_JUSTWORKS, PM_LEGACY_PASSKEY]:
            keys = []
            if self.__kd_enc_key:
                keys.append('ltk, ediv, rand')
                self.__dist_ltk = True
                self.__dist_ediv_rand = True
            if self.__kd_id_key:
                keys.append('irk')
                self.__dist_irk = True
            if self.__kd_sign_key:
                keys.append('csrk')
                self.__dist_csrk = True
            logger.debug("Set distribute key for peer: %s", ",".join(keys))
        elif self.__pairing_method in [PM_LESC_JUSTWORKS, PM_LESC_NUMCOMP, PM_LESC_OOB]:
            keys=[]
            if self.__kd_id_key:
                keys.append('irk')
                self.__dist_irk = True
            if self.__kd_sign_key:
                keys.append('csrk')
                self.__dist_csrk = True
            if self.__kd_link_key:
                keys.append('ltk')
                self.__dist_ltk = True
            logger.debug("Set distribute key for peer: %s", ",".join(keys))

    def get_key_distribution(self):
        """Get the key distribution byte from the peer.

        This method is used to convert the internal key distribution model
        into an integer value that may be used in a Pairing Request or Pairing Response.

        :return: Key distribution byte
        :rtype: int
        """
        kd = 0x00
        if self.__kd_enc_key:
            kd |= 0x01
        if self.__kd_id_key:
            kd |= 0x02
        if self.__kd_sign_key:
            kd |= 0x04
        if self.__kd_link_key:
            kd |= 0x08
        return kd

    def indicate_ltk_distribution(self, ltk):
        self.__distributed_ltk = ltk

    def indicate_rand_ediv_distribution(self, rand, ediv):
        self.__distributed_rand = rand
        self.__distributed_ediv = ediv

    def indicate_irk_distribution(self, irk):
        self.__distributed_irk = irk

    def indicate_csrk_distribution(self, csrk):
        self.__distributed_csrk = csrk

    def indicate_address_distribution(self, address, address_type):
        self.__distributed_address = address
        self.__distributed_address_type = address_type

    def is_key_distribution_complete(self):
        """
        Indicate if all keys have been distributed.
        """
        if self.__kd_enc_key and self.__distributed_ltk is None:
            return False
        if self.__kd_enc_key and (self.__distributed_rand is None or self.__distributed_ediv is None):
            return False
        if self.__kd_id_key and (self.__distributed_irk is None):
            return False
        if self.__kd_id_key and (self.__distributed_address is None and self.__distributed_address_type is None):
            return False
        if self.__kd_sign_key and (self.__distributed_csrk is None):
            return False
        return True

    @property
    def ltk(self):
        return self.__distributed_ltk

    @property
    def random(self):
        return self.__distributed_rand

    @property
    def ediv(self):
        return self.__distributed_ediv

    @property
    def irk(self):
        return self.__distributed_irk

    @property
    def csrk(self):
        return self.__distributed_csrk

    def must_dist_ltk(self):
        return self.__dist_ltk

    def must_dist_ediv_rand(self):
        return self.__dist_ediv_rand

    def must_dist_irk(self):
        return self.__dist_irk

    def must_dist_csrk(self):
        return self.__dist_csrk

    @property
    def max_key_size(self):
        return self.__max_key_size

    ##
    # Pairing method
    ##

    def set_pairing_method(self, pairing_method):
        """Set pairing method

        :param int pairing_method: Pairing method as defined in PM_* in BleSMP.
        """
        self.__pairing_method = pairing_method
        logger.info("Pairing method set to %d", self.__pairing_method)

    @property
    def pairing_method(self):
        return self.__pairing_method


    ##
    # Crypto getters/setters/generators
    ##

    def generate_legacy_rand(self):
        """Generate a Rand value based on [Vol 3] Part H. Section 2.3.5.5
        """
        # Rand
        self.__rand = generate_random_value(128)
        logger.info("(%s) Generated RAND value: %s", self.__address, self.__rand.hex())

    @property
    def rand(self):
        return self.__rand

    @rand.setter
    def rand(self, value):
        """RAND must be 128-bit long byte array
        """
        if isinstance(value, bytes) and len(value) == 16:
            self.__rand = value
            logger.debug("(%s) Set RAND value: %s", self.__address, value.hex())
        else:
            raise SMInvalidParameterFormat()

    @property
    def confirm(self):
        return self.__confirm

    @confirm.setter
    def confirm(self, value):
        """CONFIRM must be 128-bit long byte array
        """
        if isinstance(value, bytes) and len(value) == 16:
            self.__confirm = value
            logger.debug("(%s) Set confirm value: %s", self.__address, value.hex())
        else:
            raise SMInvalidParameterFormat()

    def check_peer_confirm(self, tk, preq, pres, peer, initiator=True):
        """Check peer confirm value
        """
        logger.debug("(%s) RAND value: %s", self.__address, self.__rand.hex())
        logger.debug("(%s) CONFIRM value: %s", self.__address, self.__confirm.hex())
        # First, compute confirm based on rand
        if initiator:
            _confirm = self.get_custom_function("compute_legacy_confirm_value")(
                tk,
                preq,
                pres,
                self.address,
                self.address_type,
                peer.address,
                peer.address_type,
            )
        else:
            _confirm = self.get_custom_function("compute_legacy_confirm_value")(
                tk,
                preq,
                pres,
                peer.address,
                peer.address_type,
                self.address,
                self.address_type
            )
        return _confirm == self.get_confirm_value()




    def compute_legacy_confirm_value(self, tk, preq, pres, init_addr, init_addr_type, resp_addr, resp_addr_type):
        _confirm = c1(
            tk,
            self.__rand[::-1],
            pres[::-1],
            preq[::-1],
            pack('<B', init_addr_type),
            init_addr[::-1],
            pack('<B', resp_addr_type),
            resp_addr[::-1]
        )
        logger.info("(%s) Using RAND to compute confirm: %s", self.__address, self.__rand.hex())
        logger.info("(%s) Computed CONFIRM: %s", self.__address, _confirm.hex())
        return _confirm


class SecurityManagerState(LayerState):

    STATE_IDLE = 0x00
    STATE_PAIRING_REQ = 0x01
    STATE_PAIRING_RSP = 0x02
    STATE_LEGACY_PAIRING_CONFIRM_SENT = 0x03
    STATE_LEGACY_PAIRING_CONFIRM_RECVD = 0x04
    STATE_LEGACY_PAIRING_RANDOM_SENT = 0x05
    STATE_LEGACY_PAIRING_RANDOM_RECVD = 0x06
    STATE_LESC_PUBKEY_SENT = 0x07
    STATE_LESC_PUBKEY_RECVD = 0x08
    STATE_LESC_PAIRING_CONFIRM_SENT = 0x09
    STATE_LESC_PAIRING_RANDOM_SENT = 0x0A
    STATE_LESC_PAIRING_RANDOM_RECVD = 0x0B
    STATE_LESC_DHK_CHECK_SENT = 0x0C
    STATE_LESC_DHK_CHECK_RECVD = 0x0D
    STATE_PAIRING_DONE = 0x0E
    STATE_DISTRIBUTE_KEY = 0x0F
    STATE_LESC_PUBKEY_RECVD_STAGEN = 0x10
    STATE_LESC_PAIRING_CONFIRM_SENT_STAGEN = 0x11
    STATE_LESC_PAIRING_RANDOM_SENT_STAGEN = 0x12
    STATE_PAIRING_DONE = 0xFF

    def __init__(self):
        super().__init__()

        # Peers' states
        self.initiator = None
        self.responder = None

        # Current state
        self.state = SecurityManagerState.STATE_IDLE


        # Pairing material
        self.pairing_req = None
        self.pairing_resp = None
        self.tk = b'\x00'*16
        self.stk = b'\x00'*16
        self.mackey = None

        self.ltk = None
        self.rand = None
        self.ediv = None
        self.irk = None
        self.csrk = None

        # Initiator role
        self.enc_initiator = False

        # LE Secure Connections
        self.private_key = None
        self.public_key = None

        self.peer_public_key = None
        self.shared_secret = None

        # Method
        self.method = None

        # Passkey counter
        self.passkey_value = None
        self.passkey_counter = 1

        # Security database in use
        self.database = None

@alias('smp')
@state(SecurityManagerState)
class SMPLayer(Layer):
    def configure(self, options={}):
        # SM packets dispatch
        self.__handlers = {
            SM_PAIRING_REQUEST : self.on_pairing_request,
            SM_PAIRING_RESPONSE : self.on_pairing_response,
            SM_CONFIRM : self.on_pairing_confirm,
            SM_RANDOM : self.on_pairing_random,
            SM_FAILED : self.on_pairing_failed,
            SM_ENCRYPTION_INFORMATION : self.on_encryption_information,
            SM_MASTER_IDENTIFICATION : self.on_master_identification,
            SM_IDENTITY_INFORMATION : self.on_identity_information,
            SM_IDENTITY_ADDRESS_INFORMATION : self.on_identity_address_information,
            SM_SIGNING_INFORMATION : self.on_signing_information,
            SM_SECURITY_REQUEST : self.on_security_request,
            SM_PUBLIC_KEY : self.on_public_key,
            SM_DHKEY_CHECK : self.on_dhkey_check,
        }

        # default parameters
        self.__parameters = Pairing()

    @property
    def pairing_parameters(self):
        return self.__parameters

    @pairing_parameters.setter
    def pairing_parameters(self, value):
        self.__parameters = value

    def get_custom_function(self, function_name):
        if (
            self.__parameters is not None and
            self.__parameters.custom_functions is not None and
            getattr(self.__parameters.custom_functions, function_name) is not None and
            callable(getattr(self.__parameters.custom_functions, function_name))
        ):
            return getattr(self.__parameters.custom_functions, function_name)
        else:
            return getattr(self, function_name)

    ##########
    # Helpers
    ##########

    def is_pairing_authenticated(self):
        """Indicates if the pairing uses a pairing method enforcing authentication.
        """
        return self.state.method in AUTHENTICATED_METHODS

    def key_generation_method_selection(self, initiator, responder):
        """
        This method returns the key generation method to select according to
        the exchanged initiator and responder parameters.
        """
        use_lesc = False
        # If initiator and responder supports LE Secure Connections,
        # use Table 2.7, Vol. 3, Part H, Bluetooth Core Specification v5.3, p. 1573
        if initiator.support_lesc() and responder.support_lesc():
            use_lesc = True
            if initiator.support_oob() or responder.support_oob():
                return PM_LESC_OOB
            elif not initiator.support_mitm() and not responder.support_mitm():
                return PM_LESC_JUSTWORKS
        # if at least one of the device does not support LE Secure Connections,
        # use Table 2.6, Vol. 3, Part H, Bluetooth Core Specification v5.3, p. 1572
        else:
            use_lesc = False
            if initiator.support_oob() and responder.support_oob():
                return PM_LEGACY_OOB
            elif not initiator.support_mitm() and not responder.support_mitm():
                return PM_LEGACY_JUSTWORKS

        # If we reach this point, we need to define pairing according to IO capabilities
        # (see Table 2.8, Vol. 3, Part H, Bluetooth Core Specification v5.3, p. 1573)
        try:
            return IOCAP_KEY_GENERATION_MAPPING[(initiator.iocap, responder.iocap)][int(use_lesc)]
        except IndexError:
            # it looks like an error occurred, let's return None
            return None

    def reset_state(self):
        """Reset the values in the state.
        """

        # General state
        self.state.state = SecurityManagerState.STATE_IDLE

        # Peers' states
        self.state.initiator = None
        self.state.responder = None

        # Pairing material
        self.state.pairing_req = None
        self.state.pairing_resp = None
        self.state.tk = b'\x00'*16
        self.state.stk = b'\x00'*16
        self.state.mackey = None

        self.state.ltk = None
        self.state.rand = None
        self.state.ediv = None
        self.state.irk = None
        self.state.csrk = None

        # Initiator role
        self.state.enc_initiator = False

        # LE Secure Connections
        self.state.private_key = None
        self.state.public_key = None

        self.state.peer_public_key = None
        self.state.shared_secret = None

        # Method
        self.state.method = None
        self.state.last_failure = None

    def set_security_database(self, database):
        self.state.database = database

    @property
    def security_database(self):
        return self.state.database

    def set_initiator_role(self):
        self.state.enc_initiator = True

    def set_responder_role(self):
        self.state.enc_initiator = False

    def is_initiator(self):
        return self.state.enc_initiator

    def check_initiator_legacy_confirm(self, tk):
        """Check initiator peer confirm value given a TK and the corresponding random value.

        :param SM_Peer: Peer to check
        :param bytes tk: Temporary Key
        :param Packet preq: Pairing request
        :param Packet pres: Pairing response
        :param SM_Peer initiator: Pairing initiator
        :param SM_Peer responder: Pairing responder
        """
        logger.debug("[check_initiator_legacy_confirm] RAND=%s", self.state.initiator.rand.hex())
        # Compute expected confirm value
        expected_confirm = self.compute_legacy_confirm_value(
            tk,
            self.state.initiator.rand,
            self.state.pairing_req,
            self.state.pairing_resp,
            self.state.initiator,
            self.state.responder
        )
        logger.debug("[check_initiator_legacy_confirm] Computed CONFIRM=%s", expected_confirm.hex())
        logger.debug("[check_initiator_legacy_confirm] Expected CONFIRM=%s", self.state.initiator.confirm.hex())

        # Compare with confirm value
        return (expected_confirm == self.state.initiator.confirm)

    def check_responder_legacy_confirm(self, tk, preq, pres, initiator: BDAddress, responder: BDAddress):
        """Check responder peer confirm value given a TK and the corresponding random value.

        :param SM_Peer: Peer to check
        :param bytes tk: Temporary Key
        :param Packet preq: Pairing request
        :param Packet pres: Pairing response
        :param SM_Peer initiator: Pairing initiator
        :param SM_Peer responder: Pairing responder
        """
        logger.debug("[check_responder_legacy_confirm] RAND=%s", self.state.initiator.rand.hex())

        # Compute expected confirm value
        expected_confirm = self.compute_legacy_confirm_value(
            tk,
            self.state.responder.rand,
            self.state.pairing_req,
            self.state.pairing_resp,
            self.state.initiator,
            self.state.responder
        )

        logger.debug("[check_initiator_legacy_confirm] Computed CONFIRM=%s", expected_confirm.hex())
        logger.debug("[check_initiator_legacy_confirm] Expected CONFIRM=%s", self.state.responder.confirm.hex())

        # Compare with confirm value
        return (expected_confirm == self.state.responder.confirm)

    def compute_exchange_value(self, mackey, initiator, responder, r, iocap):
        logger.debug("compute_exchange_value: %s, %s", initiator.address, responder.address)
        e = f6(
            mackey,
            initiator.rand,
            responder.rand,
            b"\x00"*12+pack('>I', r),
            iocap,
            pack('<B', initiator.address_type) +
            initiator.address[::-1],
            pack('<B', responder.address_type) +
            responder.address[::-1]
        )

        return e

    def compute_ltk_and_mackey(self, shared_secret, initiator, responder):
        logger.debug("compute_ltk_and_mackey: %s, %s", initiator.address, responder.address)
        output = f5(
            shared_secret,
            initiator.rand,
            responder.rand,
            pack('<B', initiator.address_type) +
            initiator.address[::-1],
            pack('<B', responder.address_type) +
            responder.address[::-1]
        )

        mac_key, ltk = output[:16], output[16:] # 16 MSB -> macKey, 16 LSB -> LTK
        return mac_key, ltk

    def compute_lesc_numeric_comparison(self, initiator_public_key, responder_public_key, initiator_rand, responder_rand):
        _value = g2(
            bytes.fromhex("{:064x}".format(initiator_public_key.public_numbers().x)),
            bytes.fromhex("{:064x}".format(responder_public_key.public_numbers().x)),
            initiator_rand,
            responder_rand
        )
        cropped_digits = unpack(">I", _value)[0] % (10**6)
        return cropped_digits

    def check_lesc_confirm_value(self, initiator_public_key, responder_public_key, random_number, r, value):
        _expected_value = self.compute_lesc_confirm_value(
            initiator_public_key,
            responder_public_key,
            random_number,
            r
        )
        return _expected_value == value

    def compute_lesc_confirm_value(self, initiator_public_key, responder_public_key, random_number, r):
        _confirm = f4(
            bytes.fromhex("{:064x}".format(responder_public_key.public_numbers().x)),
            bytes.fromhex("{:064x}".format(initiator_public_key.public_numbers().x)),
            random_number,
            r
        )
        logger.info("Computed LESC CONFIRM: %s", _confirm.hex())
        return _confirm

    def compute_legacy_confirm_value(self, tk, rand, preq, pres, initiator, responder):
        """Compute Confirm value as described in [Vol 3] Part H, Section 2.3.5.5

        This value is not ready to be set in a SM_Confirm packet as-is, it needs
        to be byte-reversed to be correctly decoded.

        :param bytes tk: Temporary Key
        :param bytes rand: Random to encrypt
        :param Packet preq: Pairing request
        :param Packet pres: Pairing response
        :param SM_Peer initiator: Pairing initiator
        :param SM_Peer responder: Pairing responder

        :return: Confirm value
        :rtype: bytes
        """
        logger.debug(
            "TK=%s RAND=%s, PRES=%s PREQ=%s INITA_TYPE=%02x INITA=%s RESPA_TYPE=%02x RESPA=%s",
            tk.hex(),
            rand.hex(),
            bytes(SM_Hdr()/pres)[::-1].hex(),
            bytes(SM_Hdr()/preq)[::-1].hex(),
            initiator.address_type,
            initiator.address[::-1].hex(),
            responder.address_type,
            responder.address[::-1].hex(),
        )

        # Compute the confirm value for the provided parameters
        # We need to:
        # - convert `preq` to bytes in reverse order including SM_Hdr
        # - convert `pres` to bytes in reverse order including SM_Hdr
        # - reverse order of BD addresses
        # - pack address types as 8-bit data (prefixed by 7 zeroes)

        _confirm = c1(
            tk,
            rand,
            bytes(SM_Hdr()/pres)[::-1],
            bytes(SM_Hdr()/preq)[::-1],
            pack('<B', initiator.address_type),
            initiator.address[::-1],
            pack('<B', responder.address_type),
            responder.address[::-1]
        )
        return _confirm



    def check_lesc_numeric_comparison(self, value):
        """Ask the user to check if the numeric comparison is valid.
        """
        print("Computed value is: ", value)
        print("y to continue process, n to discard")
        user_input = input()
        return user_input.upper().startswith("Y")

    def generate_p256_keypair(self):
        """Generate the P256 private key / public key
        """
        return generate_p256_keypair()

    def generate_legacy_ltk(self):
        """Generate the LTK used in legacy pairing.
        """
        if self.is_initiator():
            return generate_random_value(8*self.state.initiator.max_key_size)
        else:
            return generate_random_value(8*self.state.responder.max_key_size)

    def generate_legacy_random(self):
        """Generate the rand used in legacy pairing.
        """
        return generate_random_value(8*8)

    def generate_legacy_ediv(self):
        """Generate the EDIV used in legacy pairing.
        """
        return randint(0, 0x10000)

    def generate_irk(self):
        """Generate the IRK used in legacy and secure pairing.
        """
        if self.is_initiator():
            return generate_random_value(8*self.state.initiator.max_key_size)
        else:
            return generate_random_value(8*self.state.responder.max_key_size)

    def generate_csrk(self):
        """Generate the CSRK used in legacy and secure pairing.
        """
        if self.is_initiator():
            return generate_random_value(8*self.state.initiator.max_key_size)
        else:
            return generate_random_value(8*self.state.responder.max_key_size)

    def get_passkey_entry(self):
        """Ask the user to enter the passkey entry.
        """
        print("Enter passkey entry: ")
        passkey_entry = input()
        return int(passkey_entry)

    def get_pin_code(self):
        """Ask the user to enter the PIN code (a.k.a. Temporary Key)
        """
        self_iocap = self.state.initiator.iocap if self.is_initiator() else self.state.responder.iocap

        if self_iocap == IOCAP_KEYBD_ONLY:
            print("Enter pin code: ")
            pin_code = input()
            return int(pin_code)
        else:
            pin_code = randint(0, 999999)
            print("Randomly generated pin code: ", pin_code)
            return pin_code

    ##########################################
    # Incoming requests and responses
    ##########################################

    @instance('l2cap')
    def on_packet(self, instance, smp_pkt):
        """SMP packet reception callback

        This method dispatches every LE SMP packet received.

        :param Packet packet: Scapy packet containing SMP material
        """

        if smp_pkt.sm_command in self.__handlers:
            self.__handlers[int(smp_pkt.sm_command)](smp_pkt.getlayer(1))

    def request_pairing(self, parameters=None):
        if self.state.state in (SecurityManagerState.STATE_IDLE, SecurityManagerState.STATE_PAIRING_DONE):

            self.state.last_failure = None
            self.reset_state()
            if parameters is not None and isinstance(parameters, Pairing):
                self.pairing_parameters = parameters

            # We are the responder
            self.state.enc_initiator = False

            # Build authentication flag according to the parameters
            authentication_flag = 0x00
            if self.pairing_parameters.bonding:
                authentication_flag |= 0x01
            if self.pairing_parameters.mitm:
                authentication_flag |= 0x04
            if self.pairing_parameters.lesc:
                authentication_flag |= 0x08
            if self.pairing_parameters.keypress:
                authentication_flag |= 0x10
            if self.pairing_parameters.ct2:
                authentication_flag |= 0x20


            self.send_data(
                SM_Security_Request(
                    authentication = authentication_flag
                )
            )
            return True

        else:
            logger.info('We are in an inconsistent state.')
            return False


    def initiate_pairing(self, parameters=None):
        """
        Initiate a pairing procedure.
        """
        if self.state.state in (SecurityManagerState.STATE_IDLE, SecurityManagerState.STATE_PAIRING_DONE) :
            if parameters is not None and isinstance(parameters, Pairing):
                self.pairing_parameters = parameters

            self.state.last_failure = None
            self.reset_state()

            logger.info('Pairing Request initiation ...')

            # Get the current connection handle
            conn_handle = self.get_layer('l2cap').state.conn_handle

            # Get the current link layer state
            local_conn = self.get_layer('ll').state.get_connection(conn_handle)

            # Get the local and remote addresses values and types
            local_peer_addr = local_conn['local_peer_addr']
            local_peer_addr_type = local_conn['local_peer_addr_type']
            local_addr_object = BDAddress.from_bytes(
                local_peer_addr,
                addr_type = BDAddress.PUBLIC if
                            local_peer_addr_type == 0 else
                            BDAddress.RANDOM
            )

            # We are the initiator
            self.state.enc_initiator = True

            # Create the responder SM_Peer instance
            # (along with all its parameters are defined in the pairing request)
            self.state.initiator = SM_Peer(local_addr_object)

            self.state.initiator.set_security_parameters(
                oob=self.pairing_parameters.oob,
                bonding=self.pairing_parameters.bonding,
                mitm=self.pairing_parameters.mitm,
                lesc=self.pairing_parameters.lesc,
                keypress=self.pairing_parameters.keypress,
                max_key_size = self.pairing_parameters.max_key_size
            )

            self.state.initiator.iocap = self.pairing_parameters.iocap

            # Store initiator key distribution options
            self.state.initiator.distribute_keys(
                enc_key = self.pairing_parameters.enc_key,
                id_key = self.pairing_parameters.id_key,
                sign_key = self.pairing_parameters.sign_key,
                link_key = self.pairing_parameters.link_key
            )

            # Send our pairing response
            pairing_req = SM_Pairing_Request(
                iocap=self.state.initiator.iocap,
                oob=self.state.initiator.oob,
                authentication=self.state.initiator.authentication,
                max_key_size=self.state.initiator.max_key_size,
                initiator_key_distribution=self.state.initiator.get_key_distribution(),
                responder_key_distribution=self.state.initiator.get_key_distribution()
            )

            # Save pairing request
            self.state.pairing_req = pairing_req


            # Update current state
            self.state.state = SecurityManagerState.STATE_PAIRING_RSP

            self.send_data(pairing_req)
            return True
        else:
            logger.info('We are in an inconsistent state.')
            return False

    def on_public_key(self, public_key_pkt):
        """Method called when a public key is received.

        :param SM_Public_Key public_key_pkt: Public Key packet
        """



        # We received the public key, now transmit our own
        if self.state.state == SecurityManagerState.STATE_PAIRING_REQ:

            # Extract X and Y from public key packet
            x = int(public_key_pkt.key_x[::-1].hex(), 16)
            y = int(public_key_pkt.key_y[::-1].hex(), 16)

            # We can now generate the ECDH shared secret
            self.state.peer_public_key = generate_public_key_from_coordinates(x, y)
            self.state.shared_secret = generate_diffie_hellman_shared_secret(self.state.private_key, self.state.peer_public_key)

            own_x = bytes.fromhex("{:064x}".format(self.state.public_key.public_numbers().x))[::-1]
            own_y = bytes.fromhex("{:064x}".format(self.state.public_key.public_numbers().y))[::-1]

            self.send_data(
                SM_Public_Key(
                    key_x = own_x,
                    key_y = own_y
                )
            )

            # Generate a RAND and compute CONFIRM
            self.state.responder.generate_legacy_rand()

            if self.state.method != PM_LESC_PASSKEY:
                self.state.responder.confirm = self.get_custom_function("compute_lesc_confirm_value")(
                    self.state.peer_public_key,  # we are the responder, peer is the initiator
                    self.state.public_key,  # set our own public key
                    self.state.responder.rand,
                    b"\x00" # r equals 0 in JustWorks and NumComp
                )

                logger.debug("[send_pairing_confirm] Computed CONFIRM=%s", self.state.responder.confirm.hex())

                # Update current state
                self.state.state = SecurityManagerState.STATE_LESC_PAIRING_CONFIRM_SENT

                self.send_data(
                    SM_Confirm(
                        confirm=self.state.responder.confirm[::-1]
                    )
                )
            else:

                # Set counter to 1
                self.state.passkey_counter = 1

                self.state.state = SecurityManagerState.STATE_LESC_PUBKEY_RECVD_STAGEN

        elif self.state.state == SecurityManagerState.STATE_LESC_PUBKEY_SENT:

            # Extract X and Y from public key packet
            x = int(public_key_pkt.key_x[::-1].hex(), 16)
            y = int(public_key_pkt.key_y[::-1].hex(), 16)

            # We can now generate the ECDH shared secret
            self.state.peer_public_key = generate_public_key_from_coordinates(x, y)
            self.state.shared_secret = generate_diffie_hellman_shared_secret(self.state.private_key, self.state.peer_public_key)

            if self.state.method != PM_LESC_PASSKEY:
                self.state.state = SecurityManagerState.STATE_LESC_PUBKEY_RECVD
            else:
                # Set counter to 1
                self.state.passkey_counter = 1
                # Collect passkey entry
                self.state.passkey_value = self.get_custom_function("get_passkey_entry")()

                # Generate Nai
                self.state.initiator.generate_legacy_rand()

                self.state.initiator.confirm = self.get_custom_function("compute_lesc_confirm_value")(
                    self.state.peer_public_key,  # set our own public key
                    self.state.public_key,  # we are the initiator, peer is the responder
                    self.state.initiator.rand,
                    bytes([((self.state.passkey_value >> (self.state.passkey_counter - 1)) & 1) + 0x80])
                )

                self.state.state = SecurityManagerState.STATE_LESC_PAIRING_CONFIRM_SENT_STAGEN
                self.send_data(
                    SM_Confirm(
                        confirm = self.state.initiator.confirm[::-1]
                    )
                )



    def on_security_request(self, security_request):
        """Method called when a security request is received.
        """
        # Note: For now I keep this out of the set of states of SMP layer.
        logger.info("Receiving Security Request")
        if self.state.state in (SecurityManagerState.STATE_IDLE, SecurityManagerState.STATE_PAIRING_DONE):
            if self.pairing_parameters.accept_pairing:
                # We initiate a new pairing according to the security parameters
                # Note: we ignore authentication parameter for now, use our own
                self.initiate_pairing()

            else:
                logger.info('Pairing declined.')

                self.send_data(
                    SM_Failed(
                        reason = SM_ERROR_UNSUPP_PAIRING
                    )
                )

                # Return to IDLE mode
                self.reset_state()
                self.state.last_failure = SM_ERROR_UNSUPP_PAIRING
        else:
            logger.info('Unexpected packet received, report error and return to idle.')

            # Notify error
            error = SM_Failed(
                reason = SM_ERROR_UNSPEC_REASON
            )
            self.send_data(error)

            # Return to IDLE mode
            self.reset_state()
            self.state.last_failure = SM_ERROR_UNSPEC_REASON

    def on_pairing_request(self, pairing_req):
        """Method called when a pairing request is received.

        :param SM_Pairing_Request pairing_req: Pairing request packet
        """
        logger.info('Received Pairing Request')

        # Make sure we are in a state that allows this pairing request
        if self.state.state in (SecurityManagerState.STATE_IDLE, SecurityManagerState.STATE_PAIRING_DONE):
            self.state.last_failure = None
            self.reset_state()
            if self.pairing_parameters.accept_pairing:

                logger.info('Pairing Request accepted, processing ...')

                # Save pairing request
                self.state.pairing_req = pairing_req


                # Get the current connection handle
                conn_handle = self.get_layer('l2cap').state.conn_handle

                # Get the current link layer state
                local_conn = self.get_layer('ll').state.get_connection(conn_handle)

                # Get the local and remote addresses values and types
                local_peer_addr = local_conn['local_peer_addr']
                local_peer_addr_type = local_conn['local_peer_addr_type']

                local_addr_object = BDAddress.from_bytes(
                    local_peer_addr,
                    addr_type = BDAddress.PUBLIC if
                                local_peer_addr_type == 0 else
                                BDAddress.RANDOM
                )

                remote_peer_addr = local_conn['remote_peer_addr']
                remote_peer_addr_type = local_conn['remote_peer_addr_type']

                remote_addr_object = BDAddress.from_bytes(
                    remote_peer_addr,
                    addr_type = BDAddress.PUBLIC if
                                remote_peer_addr_type == 0 else
                                BDAddress.RANDOM
                )

                # We are definitely not the initiator but the responder
                self.state.enc_initiator = False
                self.state.responder = SM_Peer(local_addr_object)
                # Propagate our parameters to SM_Peer
                self.state.responder.iocap = self.pairing_parameters.iocap
                self.state.responder.set_security_parameters(
                    lesc=self.pairing_parameters.lesc,
                    mitm=self.pairing_parameters.mitm,
                    bonding=self.pairing_parameters.bonding,
                    keypress=self.pairing_parameters.keypress,
                    max_key_size=self.pairing_parameters.max_key_size,
                    oob=self.pairing_parameters.oob
                )
                self.state.responder.distribute_keys(
                    enc_key=self.pairing_parameters.enc_key,
                    id_key=self.pairing_parameters.id_key,
                    sign_key=self.pairing_parameters.sign_key,
                    link_key=self.pairing_parameters.link_key
                )
                # Create the initiator SM_Peer instance
                # (along with all its parameters are defined in the pairing request)
                self.state.initiator = SM_Peer(remote_addr_object)

                self.state.initiator.set_security_parameters(
                    oob=(pairing_req.oob == 0x01),
                    bonding=((pairing_req.authentication & 0x03) != 0),
                    mitm=((pairing_req.authentication & 0x04) != 0),
                    lesc=((pairing_req.authentication & 0x08) != 0),
                    keypress=((pairing_req.authentication & 0x10) != 0),
                    ct2=((pairing_req.authentication & 0x20) != 0),
                    max_key_size = pairing_req.max_key_size
                )
                self.state.initiator.iocap = pairing_req.iocap

                # Store initiator key distribution options
                self.state.initiator.distribute_keys(
                    enc_key = ((pairing_req.initiator_key_distribution & 0x01) != 0),
                    id_key = ((pairing_req.initiator_key_distribution & 0x02) != 0),
                    sign_key =((pairing_req.initiator_key_distribution & 0x04) != 0),
                    link_key = ((pairing_req.initiator_key_distribution & 0x08) != 0)
                )

                # Send our pairing response
                pairing_resp = SM_Pairing_Response(
                    iocap=self.state.responder.iocap,
                    oob=self.state.responder.oob,
                    authentication=self.state.responder.authentication,
                    max_key_size=self.state.responder.max_key_size,
                    initiator_key_distribution=self.state.initiator.get_key_distribution(),
                    responder_key_distribution=self.state.responder.get_key_distribution()
                )

                # Save pairing response
                self.state.pairing_resp = pairing_resp

                # Update current state
                self.state.state = SecurityManagerState.STATE_PAIRING_REQ

                self.send_data(pairing_resp)


                # Check key generation method in use
                self.state.method = self.key_generation_method_selection(self.state.initiator, self.state.responder)

                if self.state.method == PM_LEGACY_JUSTWORKS:
                    self.state.tk = b"\x00" * 16
                elif self.state.method == PM_LEGACY_PASSKEY:
                    pin = self.get_custom_function("get_pin_code")()
                    self.state.tk = bytes.fromhex("00"*12 + "{:08x}".format(pin))
                elif self.state.method in (PM_LESC_NUMCOMP, PM_LESC_JUSTWORKS, PM_LESC_PASSKEY):
                    # Generate the P256 keypair
                    self.state.private_key, self.state.public_key = self.get_custom_function("generate_p256_keypair")()

            else:
                logger.info('Pairing declined.')

                # Notify error
                error = SM_Failed(
                    reason = SM_ERROR_UNSUPP_PAIRING
                )
                self.send_data(error)

                # Return to IDLE mode
                self.reset_state()
                self.state.last_failure = SM_ERROR_UNSUPP_PAIRING

        else:
            logger.info('Unexpected packet received, report error and return to idle.')

            # Notify error
            error = SM_Failed(
                reason = SM_ERROR_UNSPEC_REASON
            )
            self.send_data(error)

            # Return to IDLE mode
            self.reset_state()
            self.state.last_failure = SM_ERROR_UNSPEC_REASON

    def on_pairing_failed(self, pairing_failed):
        """Method called when a pairing failed is received.
        """
        logger.info("Pairing failed (reason=%d)", pairing_failed.reason)
        # Return to IDLE mode
        self.reset_state()
        self.state.last_failure = pairing_failed.reason


    def on_dhkey_check(self, dhkey_check):
        """Method called when a Diffie Hellman check is received.

        :param SM_DHKey_Check dhkey_check: Diffie Hellman check packet
        """
        logger.info('Received Diffie Hellman check')

        # Make sure we are in a state that allows this diffie hellman check packet
        if self.state.state == SecurityManagerState.STATE_LESC_PAIRING_RANDOM_SENT:
            logger.info('Diffie Hellman check accepted, processing ...')
            self.state.ltk, self.state.mackey = self.get_custom_function("compute_ltk_and_mackey")(
                self.state.shared_secret,
                self.state.initiator,
                self.state.responder
            )
            self.state.ltk = self.state.ltk[::-1]

            self.state.rand, self.state.ediv = b"\x00"*8, 0
            # Indicate LTK as distributed
            self.state.initiator.indicate_ltk_distribution(self.state.ltk)
            self.state.initiator.indicate_rand_ediv_distribution(self.state.rand, self.state.ediv)
            self.state.responder.indicate_ltk_distribution(self.state.ltk)
            self.state.responder.indicate_rand_ediv_distribution(self.state.rand, self.state.ediv)

            if self.state.method != PM_LESC_PASSKEY:
                rb = 0
            else:
                rb = self.state.passkey_value

            # Let's compute EA
            ea = self.get_custom_function("compute_exchange_value")(
                self.state.mackey,
                self.state.initiator,
                self.state.responder,
                rb,
                bytes(
                    [
                        self.state.initiator.authentication,
                        self.state.initiator.oob,
                        self.state.initiator.iocap
                    ]
                )
            )
            logger.info("Computed EA: %s", ea.hex())
            logger.info("Received EA: %s", dhkey_check.dhkey_check[::-1].hex())
            # If we received a valid EA, generate EB and transmit it
            if ea == dhkey_check.dhkey_check[::-1]:
                eb = self.get_custom_function("compute_exchange_value")(
                    self.state.mackey,
                    self.state.responder,
                    self.state.initiator,
                    rb,
                    bytes(
                        [
                            self.state.responder.authentication,
                            self.state.responder.oob,
                            self.state.responder.iocap
                        ]
                    )
                )


                self.state.state = SecurityManagerState.STATE_LESC_DHK_CHECK_SENT

                self.send_data(
                    SM_DHKey_Check(
                        dhkey_check=eb[::-1]
                    )
                )

                # Get the current connection handle
                conn_handle = self.get_layer('l2cap').state.conn_handle

                self.get_layer('ll').state.register_encryption_key(conn_handle, self.state.ltk[::-1])

            else:  # otherwise, fail.
                logger.info('Invalid exchange value received, report error and return to idle.')

                # Notify error
                error = SM_Failed(
                    reason = SM_ERROR_DHKEY_CHECK_FAILED
                )
                self.send_data(error)

                # Return to IDLE mode
                self.reset_state()
                self.state.last_failure = SM_ERROR_DHKEY_CHECK_FAILED

        elif self.state.state == SecurityManagerState.STATE_LESC_DHK_CHECK_SENT:

            if self.state.method == PM_LESC_PASSKEY:
                rb = self.state.passkey_value
            else:
                rb = 0

            # Compute EB
            eb = self.get_custom_function("compute_exchange_value")(
                self.state.mackey,
                self.state.responder,
                self.state.initiator,
                rb,
                bytes(
                    [
                        self.state.responder.authentication,
                        self.state.responder.oob,
                        self.state.responder.iocap
                    ]
                )
            )
            # Compare it with the received data
            logger.info("Computed EB: %s", eb.hex())
            logger.info("Received EB: %s", dhkey_check.dhkey_check[::-1].hex())

            if eb == dhkey_check.dhkey_check[::-1]:

                # Get the current connection handle
                conn_handle = self.get_layer('l2cap').state.conn_handle

                # Start encryption
                self.get_layer('ll').state.register_encryption_key(conn_handle, self.state.ltk[::-1])
                self.get_layer('ll').start_encryption(conn_handle, 0, 0)

                self.state.state = SecurityManagerState.STATE_LESC_DHK_CHECK_RECVD

            # otherwise, fail.
            else:
                logger.info('Invalid exchange value received, report error and return to idle.')

                # Notify error
                error = SM_Failed(
                    reason = SM_ERROR_DHKEY_CHECK_FAILED
                )
                self.send_data(error)

                # Return to IDLE mode
                self.reset_state()
                self.state.last_failure = SM_ERROR_DHKEY_CHECK_FAILED
        else:
            logger.info('Unexpected packet received, report error and return to idle.')

            # Notify error
            error = SM_Failed(
                reason = SM_ERROR_UNSPEC_REASON
            )
            self.send_data(error)

            # Return to IDLE mode
            self.reset_state()
            self.state.last_failure = SM_ERROR_UNSPEC_REASON


    def on_pairing_response(self, pairing_resp):
        """Method called when a pairing response is received.

        :param SM_Pairing_Response pairing_resp: Pairing response packet
        """
        logger.info('Received Pairing Response')

        # Make sure we are in a state that allows this pairing request
        if self.state.state == SecurityManagerState.STATE_PAIRING_RSP:
            logger.info('Pairing Response accepted, processing ...')

            # Save pairing response
            self.state.pairing_resp = pairing_resp


            # Get the current connection handle
            conn_handle = self.get_layer('l2cap').state.conn_handle

            # Get the current link layer state
            local_conn = self.get_layer('ll').state.get_connection(conn_handle)

            remote_peer_addr = local_conn['remote_peer_addr']
            remote_peer_addr_type = local_conn['remote_peer_addr_type']

            remote_addr_object = BDAddress.from_bytes(
                remote_peer_addr,
                addr_type = BDAddress.PUBLIC if
                            remote_peer_addr_type == 0 else
                            BDAddress.RANDOM
            )

            #Configure the responder
            self.state.responder = SM_Peer(remote_addr_object)

            self.state.responder.set_security_parameters(
                oob=(pairing_resp.oob == 0x01),
                bonding=((pairing_resp.authentication & 0x03) != 0),
                mitm=((pairing_resp.authentication & 0x04) != 0),
                lesc=((pairing_resp.authentication & 0x08) != 0),
                keypress=((pairing_resp.authentication & 0x10) != 0),
                ct2=((pairing_resp.authentication & 0x20) != 0),
                max_key_size = pairing_resp.max_key_size
            )
            self.state.responder.iocap = pairing_resp.iocap

            # Store responder key distribution options
            self.state.responder.distribute_keys(
                enc_key = ((pairing_resp.responder_key_distribution & 0x01) != 0),
                id_key = ((pairing_resp.responder_key_distribution & 0x02) != 0),
                sign_key =((pairing_resp.responder_key_distribution & 0x04) != 0),
                link_key = ((pairing_resp.responder_key_distribution & 0x08) != 0)
            )


            # Check key generation method in use
            self.state.method = self.key_generation_method_selection(self.state.initiator, self.state.responder)

            if self.state.method == PM_LEGACY_JUSTWORKS:
                self.state.tk = b"\x00" * 16
            elif self.state.method == PM_LEGACY_PASSKEY:
                pin = self.get_custom_function("get_pin_code")()
                self.state.tk = bytes.fromhex("00"*12 + "{:08x}".format(pin))
            elif self.state.method in (PM_LESC_NUMCOMP, PM_LESC_JUSTWORKS, PM_LESC_PASSKEY):
                # Generate the P256 keypair
                self.state.private_key, self.state.public_key = self.get_custom_function("generate_p256_keypair")()

            if self.state.method in (PM_LEGACY_JUSTWORKS, PM_LEGACY_PASSKEY):
                # Generate a RAND and compute CONFIRM
                self.state.initiator.generate_legacy_rand()
                self.state.initiator.confirm = self.get_custom_function("compute_legacy_confirm_value")(
                    self.state.tk,
                    self.state.initiator.rand,
                    self.state.pairing_req,
                    self.state.pairing_resp,
                    self.state.initiator,
                    self.state.responder
                )
                logger.debug("[send_pairing_confirm] Computed CONFIRM=%s", self.state.initiator.confirm.hex())

                # Send CONFIRM value (again, we need to reverse its bytes)
                confirm_value = SM_Confirm(
                    confirm = self.state.initiator.confirm[::-1]
                )
                # Update current state
                self.state.state = SecurityManagerState.STATE_LEGACY_PAIRING_CONFIRM_SENT

                self.send_data(confirm_value)

            else:

                # Let's transmit our public key
                own_x = bytes.fromhex("{:064x}".format(self.state.public_key.public_numbers().x))[::-1]
                own_y = bytes.fromhex("{:064x}".format(self.state.public_key.public_numbers().y))[::-1]

                self.state.state = SecurityManagerState.STATE_LESC_PUBKEY_SENT

                self.send_data(
                    SM_Public_Key(
                        key_x = own_x,
                        key_y = own_y
                    )
                )

        else:
            logger.info('Unexpected packet received, report error and return to idle.')

            # Notify error
            error = SM_Failed(
                reason = SM_ERROR_UNSPEC_REASON
            )
            self.send_data(error)

            # Return to IDLE mode
            self.reset_state()
            self.state.last_failure = SM_ERROR_UNSPEC_REASON

    def on_pairing_confirm(self, confirm):
        """Method called whan a pairing confirm value is received.
        """
        # Make sure we have already sent a pairing request before
        logger.info('Received Pairing Confirm value')
        if self.state.state == SecurityManagerState.STATE_PAIRING_REQ:
            logger.info('Pairing Confirm value is expected, processing ...')

            # Store remote peer Confirm value (value is stored byte-reversed in Packet)
            self.state.initiator.confirm = confirm.confirm[::-1]

            # Generate a RAND and compute CONFIRM
            self.state.responder.generate_legacy_rand()
            self.state.responder.confirm = self.get_custom_function("compute_legacy_confirm_value")(
                self.state.tk,
                self.state.responder.rand,
                self.state.pairing_req,
                self.state.pairing_resp,
                self.state.initiator,
                self.state.responder
            )
            logger.debug("[on_pairing_confirm] Computed CONFIRM=%s", self.state.responder.confirm.hex())

            # Send CONFIRM value (again, we need to reverse its bytes)
            confirm_value = SM_Confirm(
                confirm = self.state.responder.confirm[::-1]
            )
            # Update current state
            self.state.state = SecurityManagerState.STATE_LEGACY_PAIRING_CONFIRM_SENT

            self.send_data(confirm_value)



        elif self.state.state == SecurityManagerState.STATE_LEGACY_PAIRING_CONFIRM_SENT:

            # Store remote peer Confirm value (value is stored byte-reversed in Packet)
            self.state.responder.confirm = confirm.confirm[::-1]

            # Send back our random
            rand_value = SM_Random(
                random = self.state.initiator.rand[::-1]
            )

            self.state.state = SecurityManagerState.STATE_LEGACY_PAIRING_RANDOM_SENT
            self.send_data(rand_value)

        elif self.state.state == SecurityManagerState.STATE_LESC_PUBKEY_RECVD:
            logger.info('Pairing Confirm value is expected, processing ...')

            # Store remote peer Confirm value (value is stored byte-reversed in Packet)
            self.state.responder.confirm = confirm.confirm[::-1]
            self.state.initiator.generate_legacy_rand()

            # Send back our random
            rand_value = SM_Random(
                random = self.state.initiator.rand[::-1]
            )

            # Update current state
            self.state.state = SecurityManagerState.STATE_LESC_PAIRING_RANDOM_SENT

            self.send_data(rand_value)

        elif self.state.state in (
            SecurityManagerState.STATE_LESC_PUBKEY_RECVD_STAGEN,
            SecurityManagerState.STATE_LESC_PAIRING_RANDOM_SENT_STAGEN
        ):
            logger.info("Pairing Confirm value is expected, processing ... [iteration=%d]", self.state.passkey_counter)

            if self.state.passkey_counter == 1:
                # Collect passkey entry
                self.state.passkey_value = self.get_custom_function("get_passkey_entry")()

            # Extract confirm value
            self.state.initiator.confirm = confirm.confirm[::-1]

            # Generate Nbi
            self.state.responder.generate_legacy_rand()

            self.state.responder.confirm = self.get_custom_function("compute_lesc_confirm_value")(
                self.state.peer_public_key,  # we are the responder, peer is the initiator
                self.state.public_key,  # set our own public key
                self.state.responder.rand,
                bytes([((self.state.passkey_value >> (self.state.passkey_counter - 1)) & 1) + 0x80])
            )

            self.state.state = SecurityManagerState.STATE_LESC_PAIRING_CONFIRM_SENT_STAGEN

            self.send_data(
                SM_Confirm(
                    confirm = self.state.responder.confirm[::-1]
                )
            )

        elif self.state.state == SecurityManagerState.STATE_LESC_PAIRING_CONFIRM_SENT_STAGEN:
            logger.info("Pairing Confirm value is expected, processing ... [iteration=%d]", self.state.passkey_counter)

            # Extract confirm value
            self.state.responder.confirm = confirm.confirm[::-1]

            self.state.state = SecurityManagerState.STATE_LESC_PAIRING_RANDOM_SENT_STAGEN

            # Transmit random
            self.send_data(
                SM_Random(
                    random = self.state.initiator.rand[::-1]
                )
            )

        else:
            logger.info("Pairing Confirm dropped because current state is %d", self.state.state)

            # Notify error
            error = SM_Failed(
                reason = SM_ERROR_UNSPEC_REASON
            )
            self.send_data(error)

            # Return to IDLE mode
            self.reset_state()
            self.state.last_failure = SM_ERROR_UNSPEC_REASON


    def on_pairing_random(self, random_pkt):
        """Handling random packet
        """
        logger.info('Received Pairing Random value')
        if self.state.state == SecurityManagerState.STATE_LESC_PAIRING_RANDOM_SENT:
            logger.info('Pairing Random value is expected, processing ...')

            # Save initiator RAND (reverse byte order)
            self.state.responder.rand = random_pkt.random[::-1]
            if self.get_custom_function("check_lesc_confirm_value")(
                    self.state.public_key,
                    self.state.peer_public_key,
                    self.state.responder.rand,
                    b"\x00",
                    self.state.responder.confirm
            ):
                # Confirm value is valid !
                # Now, we need to display the value for numeric comparison
                if self.state.method == PM_LESC_JUSTWORKS:
                    continue_pairing = True
                else:
                    value = self.get_custom_function("compute_lesc_numeric_comparison")(
                        self.state.public_key,
                        self.state.peer_public_key,
                        self.state.initiator.rand,
                        self.state.responder.rand
                    )

                    continue_pairing = self.get_custom_function("check_lesc_numeric_comparison")(value)

                if continue_pairing:
                    logger.info("Accepted numeric comparison, continuing...")
                    self.state.ltk, self.state.mackey = self.get_custom_function("compute_ltk_and_mackey")(
                        self.state.shared_secret,
                        self.state.initiator,
                        self.state.responder
                    )
                    self.state.ltk = self.state.ltk[::-1]
                    self.state.rand, self.state.ediv = b"\x00"*8, 0
                    # Indicate LTK as distributed
                    self.state.initiator.indicate_ltk_distribution(self.state.ltk)
                    self.state.initiator.indicate_rand_ediv_distribution(self.state.rand, self.state.ediv)
                    self.state.responder.indicate_ltk_distribution(self.state.ltk)
                    self.state.responder.indicate_rand_ediv_distribution(self.state.rand, self.state.ediv)

                    # Let's compute EA
                    ea = self.get_custom_function("compute_exchange_value")(
                        self.state.mackey,
                        self.state.initiator,
                        self.state.responder,
                        0,
                        bytes(
                            [
                                self.state.initiator.authentication,
                                self.state.initiator.oob,
                                self.state.initiator.iocap
                            ]
                        )
                    )
                    self.state.state = SecurityManagerState.STATE_LESC_DHK_CHECK_SENT
                    #Transmit EA
                    self.send_data(
                        SM_DHKey_Check(
                            dhkey_check=ea[::-1]
                        )
                    )

                else:
                    logger.info("Invalid Numeric comparison (expected %s)", str(value))

                    # Send error
                    error = SM_Failed(
                        reason = SM_ERROR_NUMCOMP_FAILED
                    )
                    self.send_data(error)

                    # Return to IDLE
                    self.reset_state()
                    self.state.last_failure = SM_ERROR_NUMCOMP_FAILED
            else:
                # Send error
                error = SM_Failed(
                    reason = SM_ERROR_CONFIRM_VALUE_FAILED
                )
                self.send_data(error)

                # Return to IDLE
                self.reset_state()
                self.state.last_failure = SM_ERROR_NUMCOMP_FAILED


        elif self.state.state == SecurityManagerState.STATE_LESC_PAIRING_CONFIRM_SENT:
            logger.info('Pairing Random value is expected, processing ...')

            # Save initiator RAND (reverse byte order)
            self.state.initiator.rand = random_pkt.random[::-1]

            # Send back our random
            rand_value = SM_Random(
                random = self.state.responder.rand[::-1]
            )
            self.send_data(rand_value)
            if self.state.method == PM_LESC_JUSTWORKS:
                continue_pairing = True
            else:
                # Now, we need to display the value for numeric comparison
                value = self.get_custom_function("compute_lesc_numeric_comparison")(
                    self.state.peer_public_key,
                    self.state.public_key,
                    self.state.initiator.rand,
                    self.state.responder.rand
                )

                continue_pairing = self.get_custom_function("check_lesc_numeric_comparison")(value)
            if continue_pairing:
                logger.info("Accepted numeric comparison, continuing...")
                self.state.state = SecurityManagerState.STATE_LESC_PAIRING_RANDOM_SENT
            else:
                logger.info("Invalid Numeric comparison (expected %s)", str(value))

                # Send error
                error = SM_Failed(
                    reason = SM_ERROR_NUMCOMP_FAILED
                )
                self.send_data(error)

                # Return to IDLE
                self.reset_state()
                self.state.last_failure = SM_ERROR_NUMCOMP_FAILED

        elif self.state.state == SecurityManagerState.STATE_LEGACY_PAIRING_CONFIRM_SENT:
            logger.info('Pairing Random value is expected, processing ...')

            # Save initiator RAND (reverse byte order)
            self.state.initiator.rand = random_pkt.random[::-1]

            if self.get_custom_function("check_initiator_legacy_confirm")(self.state.tk):
                logger.info('Initiator CONFIRM successfully verified')
                # Send back our random
                rand_value = SM_Random(
                    random = self.state.responder.rand[::-1]
                )

                # Next state
                self.state.state = SecurityManagerState.STATE_LEGACY_PAIRING_RANDOM_SENT

                self.send_data(rand_value)

                # Compute our stk
                self.state.stk = s1(
                    self.state.tk,
                    self.state.responder.rand,
                    self.state.initiator.rand
                )

                logger.debug("[on_pairing_random] STK=%s", self.state.stk.hex())

                # Notify connection that we successfully negociated STK and that
                # the corresponding material is available.

                # Get the current connection handle
                conn_handle = self.get_layer('l2cap').state.conn_handle

                self.get_layer('ll').state.register_encryption_key(conn_handle, self.state.stk)

                #self.__l2cap.connection.set_stk(self.state.stk)

            else:
                logger.info(
                    "Invalid Initiator CONFIRM value (expected %s)",
                    self.state.initiator.confirm.hex(),
                )

                # Send error
                error = SM_Failed(
                    reason = SM_ERROR_CONFIRM_VALUE_FAILED
                )
                self.send_data(error)

                # Return to IDLE
                self.reset_state()
                self.state.last_failure = SM_ERROR_CONFIRM_VALUE_FAILED

        elif self.state.state == SecurityManagerState.STATE_LESC_PAIRING_RANDOM_SENT_STAGEN:
            self.state.responder.rand = random_pkt.random[::-1]
            ri = bytes([((self.state.passkey_value >> (self.state.passkey_counter - 1)) & 1) + 0x80])

            computed_confirm = self.get_custom_function("compute_lesc_confirm_value")(
                self.state.public_key,
                self.state.peer_public_key,
                self.state.responder.rand,
                ri
            )

            if computed_confirm == self.state.responder.confirm:
                if self.state.passkey_counter == 20:
                    self.state.ltk, self.state.mackey = self.get_custom_function("compute_ltk_and_mackey")(
                        self.state.shared_secret,
                        self.state.initiator,
                        self.state.responder
                    )
                    self.state.ltk = self.state.ltk[::-1]
                    self.state.rand, self.state.ediv = b"\x00"*8, 0
                    # Indicate LTK as distributed
                    self.state.initiator.indicate_ltk_distribution(self.state.ltk)
                    self.state.initiator.indicate_rand_ediv_distribution(self.state.rand, self.state.ediv)
                    self.state.responder.indicate_ltk_distribution(self.state.ltk)
                    self.state.responder.indicate_rand_ediv_distribution(self.state.rand, self.state.ediv)

                    ra = self.state.passkey_value
                    # Let's compute EA
                    ea = self.get_custom_function("compute_exchange_value")(
                        self.state.mackey,
                        self.state.initiator,
                        self.state.responder,
                        ra,
                        bytes(
                            [
                                self.state.initiator.authentication,
                                self.state.initiator.oob,
                                self.state.initiator.iocap
                            ]
                        )
                    )

                    self.state.state = SecurityManagerState.STATE_LESC_DHK_CHECK_SENT
                    # Then transmit it
                    self.send_data(
                        SM_DHKey_Check(
                            dhkey_check = ea[::-1]
                        )
                    )

                else:
                    # Increment counter
                    self.state.passkey_counter += 1

                    # Generate Nai
                    self.state.initiator.generate_legacy_rand()

                    self.state.initiator.confirm = self.get_custom_function("compute_lesc_confirm_value")(
                        self.state.peer_public_key,  # set our own public key
                        self.state.public_key,  # we are the initiator, peer is the responder
                        self.state.initiator.rand,
                        bytes([((self.state.passkey_value >> (self.state.passkey_counter - 1)) & 1) + 0x80])
                    )

                    self.state.state = SecurityManagerState.STATE_LESC_PAIRING_CONFIRM_SENT_STAGEN

                    self.send_data(
                        SM_Confirm(
                            confirm = self.state.initiator.confirm[::-1]
                        )
                    )

            else:
                logger.info(
                    "Invalid responder CONFIRM value (expected %s)",
                    self.state.responder.confirm.hex(),
                )

                # Send error
                error = SM_Failed(
                    reason = SM_ERROR_CONFIRM_VALUE_FAILED
                )
                self.send_data(error)

                # Return to IDLE
                self.reset_state()
                self.state.last_failure = SM_ERROR_CONFIRM_VALUE_FAILED



        elif self.state.state == SecurityManagerState.STATE_LESC_PAIRING_CONFIRM_SENT_STAGEN:
            self.state.initiator.rand = random_pkt.random[::-1]

            ri = bytes([((self.state.passkey_value >> (self.state.passkey_counter - 1)) & 1) + 0x80])

            computed_confirm = self.get_custom_function("compute_lesc_confirm_value")(
                self.state.public_key,
                self.state.peer_public_key,
                self.state.initiator.rand,
                ri
            )

            if computed_confirm == self.state.initiator.confirm:

                if self.state.passkey_counter == 20:
                    self.state.state = SecurityManagerState.STATE_LESC_PAIRING_RANDOM_SENT
                else:
                    self.state.passkey_counter += 1
                    self.state.state = SecurityManagerState.STATE_LESC_PAIRING_RANDOM_SENT_STAGEN

                self.send_data(
                    SM_Random(
                        random = self.state.responder.rand[::-1]
                    )
                )
            else:
                logger.info(
                    "Invalid Initiator CONFIRM value (expected %s)",
                    self.state.initiator.confirm.hex(),
                )

                # Send error
                error = SM_Failed(
                    reason = SM_ERROR_CONFIRM_VALUE_FAILED
                )
                self.send_data(error)

                # Return to IDLE
                self.reset_state()
                self.state.last_failure = SM_ERROR_CONFIRM_VALUE_FAILED



        elif self.state.state == SecurityManagerState.STATE_LEGACY_PAIRING_RANDOM_SENT:
            logger.info('Pairing Random value is expected, processing ...')

            # Save responder RAND (reverse byte order)
            self.state.responder.rand = random_pkt.random[::-1]

            if self.get_custom_function("check_responder_legacy_confirm")(
                self.state.tk,
                self.state.pairing_req,
                self.state.pairing_resp,
                self.state.initiator,
                self.state.responder
            ):
                logger.info('Responder CONFIRM successfully verified')

                # Compute our stk
                self.state.stk = s1(
                    self.state.tk,
                    self.state.responder.rand,
                    self.state.initiator.rand
                )


                logger.debug("[on_pairing_random] STK=%s", self.state.stk.hex())

                # Next state
                self.state.state = SecurityManagerState.STATE_LEGACY_PAIRING_RANDOM_RECVD

                # Notify connection that we successfully negociated STK and that
                # the corresponding material is available.

                # Get the current connection handle
                conn_handle = self.get_layer('l2cap').state.conn_handle

                self.get_layer('ll').state.register_encryption_key(conn_handle, self.state.stk)
                self.get_layer('ll').start_encryption(conn_handle, 0, 0)

            else:
                logger.info(
                    "Invalid Responder CONFIRM value (expected %s)",
                    self.state.responder.confirm.hex(),
                )

                # Send error
                error = SM_Failed(
                    reason = SM_ERROR_CONFIRM_VALUE_FAILED
                )
                self.send_data(error)

                # Return to IDLE
                self.reset_state()
                self.state.last_failure = SM_ERROR_CONFIRM_VALUE_FAILED

        else:
            logger.info("Pairing Random dropped because current state is %d", self.state.state)

            # Notify error
            error = SM_Failed(
                reason = SM_ERROR_UNSPEC_REASON
            )
            self.send_data(error)

            # Return to IDLE mode
            self.reset_state()
            self.state.last_failure = SM_ERROR_UNSPEC_REASON

    def on_channel_encrypted(self):
        """Handling LL_START_ENC_RSP (channel successfully encrypted).

        This method is called when we successfully received and decrypted an
        encrypted LL_START_ENC_RSP packet from the remote peer.
        """
        # Previous state was STATE_LEGACY_PAIRING_RANDOM_SENT
        # since LL_ENC_REQ / LL_ENC_RSP / LL_START_ENC_REQ / LL_START_ENC_RSP
        # sequence has been handled by the link-layer manager.

        if self.state.state == SecurityManagerState.STATE_LEGACY_PAIRING_RANDOM_SENT:
            logger.info('[smp] Channel is now successfully encrypted')
            self.perform_key_distribution()

        elif self.state.state == SecurityManagerState.STATE_LEGACY_PAIRING_RANDOM_RECVD:
            logger.info('[smp] Channel is now successfully encrypted')
            if self.state.responder.is_key_distribution_complete():
                self.perform_key_distribution()

        elif self.state.state == SecurityManagerState.STATE_LESC_DHK_CHECK_SENT:
            logger.info('[smp] Channel is now successfully encrypted')
            self.perform_key_distribution()
        else:
            logger.error('[smp] Received an unexpected notification (LL_START_ENC_RSP)')

    def perform_key_distribution(self):

        if self.is_initiator():
            # If LTK is not None, it has been derivated from DHKey in LE SC
            # don't distribute LTK, only IRK and CSRK
            if self.state.ltk is None:
                self.state.ltk = self.get_custom_function("generate_legacy_ltk")()
                self.state.rand = self.get_custom_function("generate_legacy_random")()
                self.state.ediv = self.get_custom_function("generate_legacy_ediv")()

                # Perform key distribution to responder
                sleep(.5)
                if self.state.initiator.must_dist_ltk():
                    logger.info('[smp] sending generated LTK ...')
                    self.send_data(SM_Encryption_Information(
                        ltk=self.state.ltk
                    ))
                    logger.info('[smp] LTK sent.')
                if self.state.initiator.must_dist_ediv_rand():
                    logger.info('[smp] sending generated EDIV/RAND ...')
                    self.send_data(SM_Master_Identification(ediv = self.state.ediv, rand = self.state.rand))
                logger.info('[smp] EDIV/RAND sent.')
            if self.state.initiator.must_dist_irk():
                logger.info('[smp] sending generated IRK ...')
                self.state.irk = self.get_custom_function("generate_irk")()
                self.send_data(SM_Identity_Information(
                    irk = self.state.irk
                ))
                logger.info('[smp] IRK sent.')
                self.send_data(
                    SM_Identity_Address_Information(
                        atype = self.state.initiator.address_type,
                        address = self.state.initiator.address
                    )
                )
                logger.info('[smp] Address information sent.')


            if self.state.initiator.must_dist_csrk():
                logger.info('[smp] sending generated CSRK ...')
                self.state.csrk = self.get_custom_function("generate_csrk")()
                self.send_data(SM_Signing_Information(
                    csrk=self.state.csrk
                ))
                logger.info('[smp] CSRK sent.')

            self.pairing_done()

        else:
            if self.state.ltk is None:
                self.state.ltk = self.get_custom_function("generate_legacy_ltk")()
                self.state.rand = self.get_custom_function("generate_legacy_random")()
                self.state.ediv = self.get_custom_function("generate_legacy_ediv")()

                # Perform key distribution to initiator
                sleep(.5)
                if self.state.responder.must_dist_ltk():
                    logger.info('[smp] sending generated LTK ...')
                    self.send_data(SM_Encryption_Information(
                        ltk=self.state.ltk
                    ))
                    logger.info('[smp] LTK sent.')
                if self.state.responder.must_dist_ediv_rand():
                    logger.info('[smp] sending generated EDIV/RAND ...')
                    self.send_data(SM_Master_Identification(ediv = self.state.ediv, rand = self.state.rand))
                    logger.info('[smp] EDIV/RAND sent.')
            if self.state.responder.must_dist_irk():
                logger.info('[smp] sending generated IRK ...')
                self.state.irk = self.get_custom_function("generate_irk")()
                self.send_data(SM_Identity_Information(
                    irk = self.state.irk
                ))
                logger.info('[smp] IRK sent.')
                self.send_data(
                    SM_Identity_Address_Information(
                        atype = self.state.responder.address_type,
                        address = self.state.responder.address
                    )
                )
                logger.info('[smp] Address information sent.')
            if self.state.responder.must_dist_csrk():
                logger.info('[smp] sending generated CSRK ...')
                self.state.csrk = self.get_custom_function("generate_csrk")()

                self.state.state = SecurityManagerState.STATE_DISTRIBUTE_KEY

                self.send_data(SM_Signing_Information(
                    csrk=self.state.csrk
                ))
                logger.info('[smp] CSRK sent.')


    def on_encryption_information(self, encryption_information):
        if self.is_initiator():
            self.state.responder.indicate_ltk_distribution(encryption_information.ltk)
            if self.state.responder.is_key_distribution_complete():
                self.perform_key_distribution()
        else:
            self.state.initiator.indicate_ltk_distribution(encryption_information.ltk)
            if self.state.initiator.is_key_distribution_complete():
                self.pairing_done()

    def on_master_identification(self, master_identification):
        if self.is_initiator():
            self.state.responder.indicate_rand_ediv_distribution(master_identification.rand, master_identification.ediv)
            if self.state.responder.is_key_distribution_complete():
                self.perform_key_distribution()
        else:
            self.state.initiator.indicate_rand_ediv_distribution(master_identification.rand, master_identification.ediv)
            if self.state.initiator.is_key_distribution_complete():
                self.pairing_done()


    def on_identity_information(self, identity_information):
        if self.is_initiator():
            self.state.responder.indicate_irk_distribution(identity_information.irk)
            if self.state.responder.is_key_distribution_complete():
                self.perform_key_distribution()
        else:
            self.state.initiator.indicate_irk_distribution(identity_information.irk)
            if self.state.initiator.is_key_distribution_complete():
                self.pairing_done()


    def on_identity_address_information(self, identity_address_information):
        if self.is_initiator():
            self.state.responder.indicate_address_distribution(identity_address_information.address, identity_address_information.atype)
            if self.state.responder.is_key_distribution_complete():
                self.perform_key_distribution()
        else:
            self.state.initiator.indicate_address_distribution(identity_address_information.address, identity_address_information.atype)
            if self.state.initiator.is_key_distribution_complete():
                self.pairing_done()


    def on_signing_information(self, signing_information):
        if self.is_initiator():
            self.state.responder.indicate_csrk_distribution(signing_information.csrk)
            if self.state.responder.is_key_distribution_complete():
                self.perform_key_distribution()
        else:
            self.state.initiator.indicate_csrk_distribution(signing_information.csrk)
            if self.state.initiator.is_key_distribution_complete():
                self.pairing_done()

    def is_pairing_done(self):
        return self.state.state == SecurityManagerState.STATE_PAIRING_DONE

    def is_pairing_failed(self):
        return self.state.last_failure is not None

    def pairing_done(self):
        #print(self.state.database)

        logger.info("Pairing done.")
        if self.is_initiator():
            if self.state.ltk is not None:
                logger.info("Distributed LTK: %s", self.state.ltk.hex())
            if self.state.rand is not None:
                logger.info("Distributed RAND: %s", self.state.rand.hex())
            if self.state.ediv is not None:
                logger.info("Distributed EDIV: %s", hex(self.state.ediv))
            if self.state.irk is not None:
                logger.info("Distributed IRK: %s", self.state.irk.hex())
            if self.state.csrk is not None:
                logger.info("Distributed CSRK: %s", self.state.csrk.hex())

            if self.state.responder.ltk is not None:
                logger.info("Received LTK: %s",  self.state.responder.ltk.hex())
            if self.state.responder.random is not None:
                logger.info("Received RAND: %s", self.state.responder.random.hex())
            if self.state.responder.ediv is not None:
                logger.info("Received EDIV: %s", hex(self.state.responder.ediv))
            if self.state.responder.irk is not None:
                logger.info("Received IRK: %s", self.state.responder.irk.hex())
            if self.state.responder.csrk is not None:
                logger.info("Received CSRK: %s", self.state.responder.csrk.hex())

            # If bonding is enabled, we register in the security DB
            # both our own and the peer cryptographic material
            if self.pairing_parameters.bonding:
                logger.info("Saving cryptographic materials in security database")
                self.state.database.add(
                    self.state.initiator.bd_address,
                    authenticated=self.is_pairing_authenticated(),
                    ltk=(
                        self.state.ltk[::-1]
                        if self.state.ltk is not None else
                        None
                        ),
                    rand=(
                        self.state.rand[::-1]
                        if self.state.rand is not None else
                        None
                        ),
                    ediv=self.state.ediv,
                    irk=(
                        self.state.irk[::-1]
                        if self.state.irk is not None else
                        None
                        ),
                    csrk=(
                        self.state.csrk[::-1]
                        if self.state.csrk is not None else
                        None
                        )
                )
                self.state.database.add(
                    self.state.responder.bd_address,
                    authenticated=self.is_pairing_authenticated(),
                    ltk=(
                        self.state.responder.ltk[::-1]
                        if self.state.responder.ltk is not None else
                        None
                        ),
                    rand=(
                        self.state.responder.random[::-1]
                        if self.state.responder.random is not None else
                        None
                        ),
                    ediv=self.state.responder.ediv,
                    irk=(
                        self.state.responder.irk[::-1]
                        if self.state.responder.irk is not None else
                        None
                        ),
                    csrk=(
                        self.state.responder.csrk[::-1]
                        if self.state.responder.csrk is not None else
                        None
                        )
                )

        else:
            if self.state.ltk is not None:
                logger.info("Distributed LTK: %s", self.state.ltk.hex())
            if self.state.rand is not None:
                logger.info("Distributed RAND: %s", self.state.rand.hex())
            if self.state.ediv is not None:
                logger.info("Distributed EDIV: %s", hex(self.state.ediv))
            if self.state.irk is not None:
                logger.info("Distributed IRK: %s", self.state.irk.hex())
            if self.state.csrk is not None:
                logger.info("Distributed CSRK: %s", self.state.csrk.hex())

            if self.state.initiator.ltk is not None:
                logger.info("Received LTK: %s", self.state.initiator.ltk.hex())
            if self.state.initiator.random is not None:
                logger.info("Received RAND: %s", self.state.initiator.random.hex())
            if self.state.initiator.ediv is not None:
                logger.info("Received EDIV: %s", hex(self.state.initiator.ediv))
            if self.state.initiator.irk is not None:
                logger.info("Received IRK: %s", self.state.initiator.irk.hex())
            if self.state.initiator.csrk is not None:
                logger.info("Received CSRK: %s", self.state.initiator.csrk.hex())

            # If bonding is enabled, we register in the security DB
            # both our own and the peer cryptographic material
            if self.pairing_parameters.bonding:
                logger.info("Saving cryptographic materials in security database")
                self.state.database.add(
                    self.state.responder.bd_address,
                    authenticated=self.is_pairing_authenticated(),
                    ltk=(
                        self.state.ltk[::-1]
                        if self.state.ltk is not None else
                        None
                        ),
                    rand=(
                        self.state.rand[::-1]
                        if self.state.rand is not None else
                        None
                        ),
                    ediv=self.state.ediv,
                    irk=(
                        self.state.irk[::-1]
                        if self.state.irk is not None else
                        None
                        ),
                    csrk=(
                        self.state.csrk[::-1]
                        if self.state.csrk is not None else
                        None
                        )
                )
                self.state.database.add(
                    self.state.initiator.bd_address,
                    authenticated=self.is_pairing_authenticated(),
                    ltk=(
                        self.state.initiator.ltk[::-1]
                        if self.state.initiator.ltk is not None else
                        None
                        ),
                    rand=(
                        self.state.initiator.random[::-1]
                        if self.state.initiator.random is not None else
                        None
                        ),
                    ediv=self.state.initiator.ediv,
                    irk=(
                        self.state.initiator.irk[::-1]
                        if self.state.initiator.irk is not None else
                        None
                        ),
                    csrk=(
                        self.state.initiator.csrk[::-1]
                        if self.state.initiator.csrk is not None else
                        None
                        )
                )

        #print(self.state.database)
        self.state.state = SecurityManagerState.STATE_PAIRING_DONE

    def send_data(self, packet):
        """Helper transmitting payload to the lower layer with the SM header.
        """
        self.send('l2cap', SM_Hdr()/packet)



"""
RFStorm adaptation layer for WHAD.
"""
import logging
from threading import  Lock
from time import sleep, time

from usb.core import find, USBError, USBTimeoutError

from whad.exceptions import WhadDeviceNotFound, WhadDeviceNotReady, WhadDeviceAccessDenied

from whad.hub.generic.cmdresult import CommandResult
from whad.hub.esb import Commands as EsbCommands
from whad.hub.phy import Commands as PhyCommands
from whad.esb.esbaddr import ESBAddress
from whad.hub.unifying import Commands as UniCommands
from whad.hub.discovery import Capability, Domain
from whad.phy import Endianness
from whad.helpers import swap_bits

from ..device import VirtualDevice
from .constants import RFStormId, RFStormCommands, \
    RFStormDataRate, RFStormEndPoints, RFStormInternalStates, RFStormDomains

logger = logging.getLogger(__name__)

# Helpers functions
def get_rfstorm(index=0,bus=None, address=None):
    '''
    Returns a RFStorm USB object based on index or bus and address.
    '''
    devices = list(find(idVendor=RFStormId.RFSTORM_ID_VENDOR,
                        idProduct=RFStormId.RFSTORM_ID_PRODUCT,find_all=True))
    if bus is not None and address is not None:
        for device in devices:
            if device.bus == bus and device.address == address:
                return (devices.index(device), device)
        # No device found with the corresponding bus/address, return None
        return None

    try:
        return (index, devices[index])
    except IndexError:
        return None


class RfStorm(VirtualDevice):
    """RFStorm virtual device implementation.
    """

    INTERFACE_NAME = "rfstorm"

    @classmethod
    def list(cls):
        '''
        Returns a list of available RFStorm devices.
        '''
        available_devices = []
        try:
            for rfstorm in find(idVendor=RFStormId.RFSTORM_ID_VENDOR,
                                idProduct=RFStormId.RFSTORM_ID_PRODUCT,find_all=True):
                available_devices.append(RfStorm(bus=rfstorm.bus, address=rfstorm.address))
        except ValueError:
            logger.warning("Cannot access RFStorm, root privileges may be required.")

        return available_devices

    @property
    def identifier(self):
        '''
        Returns the identifier of the current device (e.g., bus + address in
        format "<bus>-<address>").
        '''
        return str(self.__rfstorm.bus)+"-"+str(self.__rfstorm.address)


    def __init__(self, index=0, bus=None, address=None):
        """
        Create device connection
        """
        device = get_rfstorm(index,bus=bus,address=address)
        if device is None:
            raise WhadDeviceNotFound

        self.__opened = False
        self.__opened_stream = False
        self.__channel = 0
        self.__address = b"\xFF\xFF\xFF\xFF\xFF"
        self.__ptx = False
        self.__scanning = False
        self.__acking = False
        self.__ack_payload = None
        self.__check_ack = False

        self.__phy_sync = None
        self.__phy_rate = RFStormDataRate.RF_2MBPS
        self.__phy_size = 32
        self.__phy_endianness = Endianness.BIG

        self.__internal_state = RFStormInternalStates.NONE
        self.__domain = RFStormDomains.RFSTORM_RAW_ESB
        _, self.__rfstorm = device
        self.__last_packet_timestamp = 0

        self.__supported_frequency_range = (2400000000, 2500000000)
        self.__lock = Lock()
        super().__init__()

    def reset(self):
        self.__rfstorm.reset()

    def open(self):
        try:
            self.__rfstorm.set_configuration()
        except USBError as err:
            if err.errno == 13:
                raise WhadDeviceAccessDenied("rfstorm") from err
            raise WhadDeviceNotReady() from err

        self._dev_id = self._get_serial_number()
        self._fw_author = self._get_manufacturer()
        self._fw_url = self._get_url()
        self._fw_version = self._get_firmware_version()
        self._dev_capabilities = self._get_capabilities()

        self.__opened_stream = False
        self.__opened = True

        # Ask parent class to run a background I/O thread
        super().open()


    def _get_serial_number(self):
        return bytes.fromhex(
            f"{self.__rfstorm.bus:02x}"*8 +
            f"{self.__rfstorm.address:02x}"*8
        )

    def _get_manufacturer(self):
        return "Marc Newlin (BastilleResearch)".encode('utf-8')


    # Discovery related functions
    def _get_capabilities(self):
        capabilities = {
            Domain.Esb : (
                (Capability.Sniff | Capability.Inject | Capability.SimulateRole \
                    | Capability.NoRawData),
                [
                    EsbCommands.Sniff,
                    EsbCommands.Send,
                    EsbCommands.Start,
                    EsbCommands.Stop,
                    EsbCommands.SetNodeAddress,
                    EsbCommands.PrimaryReceiverMode,
                    EsbCommands.PrimaryTransmitterMode
                ]
            ),
            Domain.LogitechUnifying : (
                (Capability.Sniff | Capability.Inject | Capability.SimulateRole \
                 | Capability.NoRawData),
                [
                    UniCommands.Sniff,
                    UniCommands.Send,
                    UniCommands.Start,
                    UniCommands.Stop,
                    UniCommands.SetNodeAddress,
                    UniCommands.LogitechMouseMode,
                    UniCommands.LogitechKeyboardMode,
                    UniCommands.LogitechDongleMode
                ]
            ),
            Domain.Phy: (
                (Capability.Sniff | Capability.Inject | Capability.NoRawData),
                [
                    PhyCommands.SetGFSKModulation,
                    PhyCommands.GetSupportedFrequencies,
                    PhyCommands.SetFrequency,
                    PhyCommands.SetDataRate,
                    PhyCommands.SetEndianness,
                    PhyCommands.SetPacketSize,
                    PhyCommands.SetSyncWord,
                    PhyCommands.Sniff,
                    PhyCommands.Send,
                    PhyCommands.Start,
                    PhyCommands.Stop
                ]
            ),

        }
        return capabilities

    def _get_firmware_version(self):
        return (1, 0, 0)

    def _get_url(self):
        return "https://github.com/BastilleResearch/nrf-research-firmware".encode('utf-8')


    def close(self):
        """
        Close current device.
        """
        # Ask parent class to stop I/O thread
        super().close()

        # Close underlying device.
        self.__opened = False

    def _rfstorm_send_command(self, command, data=b"", timeout=200, no_response=False):
        data = [command] + list(data)
        with self.__lock:
            try:
                self.__rfstorm.write(RFStormEndPoints.RFSTORM_COMMAND_ENDPOINT,
                                     data, timeout=timeout)
            except USBTimeoutError:
                return False
            response = self._rfstorm_read_response()

        # If we have a response, return it
        if not no_response:
            return response

        # Success
        return True

    def _rfstorm_read_response(self, timeout=200):
        try:
            return bytes(self.__rfstorm.read(RFStormEndPoints.RFSTORM_RESPONSE_ENDPOINT,
                                             64, timeout=timeout))
        except USBTimeoutError:
            return None

    def _rfstorm_check_success(self, data):
        return data is not None and len(data) > 0 and data[0] > 0

    def _rfstorm_read_packet(self):
        return self._rfstorm_send_command(RFStormCommands.RFSTORM_CMD_RECV)

    def _rfstorm_promiscuous_mode(self, prefix=b""):
        data = bytes([len(prefix)]) + prefix
        return self._rfstorm_send_command(RFStormCommands.RFSTORM_CMD_PROMISCUOUS,
                                          data, no_response=True)

    def _rfstorm_generic_promiscuous_mode(self, prefix=b"", rate=RFStormDataRate.RF_2MBPS,
                                          payload_length=32):
        data = bytes([len(prefix), rate, payload_length]) + prefix
        return self._rfstorm_send_command(RFStormCommands.RFSTORM_CMD_PROMISCUOUS_GENERIC,
                                          data, no_response=True)

    def _rfstorm_sniffer_mode(self, address=b""):
        data = bytes([len(address)]) + address
        return self._rfstorm_send_command(RFStormCommands.RFSTORM_CMD_SNIFF, data,
                                          no_response=True)

    def _rfstorm_tone_mode(self):
        return self._rfstorm_send_command(RFStormCommands.RFSTORM_CMD_TONE)

    def _rfstorm_transmit_payload(self, payload, timeout=4, retransmits=1):
        data = bytes([len(payload), timeout, retransmits]) + payload
        return self._rfstorm_check_success(
            self._rfstorm_send_command(RFStormCommands.RFSTORM_CMD_TRANSMIT, data)
        )
    def _rfstorm_transmit_payload_generic(self, payload, address=b"\x33\x33\x33\x33\x33"):
        data = bytes([len(payload), len(address)]) + payload + address
        return self._rfstorm_check_success(
            self._rfstorm_send_command(RFStormCommands.RFSTORM_CMD_TRANSMIT_GENERIC, data)
        )

    def _rfstorm_transmit_ack_payload(self, payload):
        data = bytes([len(payload)]) + payload
        return self._rfstorm_send_command(RFStormCommands.RFSTORM_CMD_TRANSMIT_ACK,
                                          data, no_response=True)

    def _rfstorm_set_channel(self, channel):
        if channel < 0 or channel > 125:
            return False

        data = bytes([channel])
        return self._rfstorm_send_command(RFStormCommands.RFSTORM_CMD_SET_CHANNEL, data)

    def _rfstorm_get_channel(self, channel):
        return self._rfstorm_send_command(RFStormCommands.RFSTORM_CMD_GET_CHANNEL)[0]

    def _rfstorm_enable_lna(self):
        return self._rfstorm_send_command(RFStormCommands.RFSTORM_CMD_ENABLE_LNA)

    def write(self, data):
        if not self.__opened:
            raise WhadDeviceNotReady()

    def read(self):
        """Read incoming data
        """
        if not self.__opened:
            raise WhadDeviceNotReady()
        if self.__opened_stream:
            if self.__domain == RFStormDomains.RFSTORM_PHY:
                try:
                    data = self._rfstorm_read_packet()
                except USBTimeoutError:
                    data = b""
                if data is not None and isinstance(data, bytes) and \
                        len(data) >= 1 and data != b"\xFF":
                    self.__last_packet_timestamp = time()
                    if len(data[:5]) >= 3:
                        if self.__phy_endianness == Endianness.LITTLE:
                            data = bytes([swap_bits(i) for i in data])
                        self._send_whad_pdu(data[5:], data[:5], int(self.__last_packet_timestamp))

            else:
                if self.__scanning:
                    if time() - self.__last_packet_timestamp > 1:
                        if self.__internal_state == RFStormInternalStates.PROMISCUOUS_SNIFFING:
                            self.__channel = (self.__channel + 1) % 100
                            self._rfstorm_set_channel(self.__channel)
                            sleep(0.05)
                        elif self.__internal_state == RFStormInternalStates.SNIFFING:
                            for i in range(0,100):
                                self._rfstorm_set_channel(i)
                                if self._rfstorm_transmit_payload(b"\x0f\x0f\x0f\x0f",1,1):
                                    self.__last_packet_timestamp = time()
                                    self.__channel = i
                                    self._send_whad_pdu(b"", address=self.__address)
                                    break

                if not self.__ptx:
                    try:
                        data = self._rfstorm_read_packet()
                    except USBTimeoutError:
                        data = b""

                    if data is not None and isinstance(data, bytes) and \
                            len(data) >= 1 and data != b"\xFF":
                        if self.__acking:
                            if self.__ack_payload is not None:
                                self._rfstorm_transmit_ack_payload(self.__ack_payload)
                                self.__ack_payload = None
                            else:
                                self._rfstorm_transmit_ack_payload(b"")
                        self.__last_packet_timestamp = time()
                        if self.__internal_state == RFStormInternalStates.PROMISCUOUS_SNIFFING:
                            if len(data[:5]) >= 3:
                                self._send_whad_pdu(data[5:], data[:5])
                        elif self.__internal_state == RFStormInternalStates.SNIFFING:
                            self._send_whad_pdu(data[1:], self.__address)
                else:
                    sleep(0.1)


    # Virtual device whad message builder
    def _send_whad_pdu(self, pdu, address, timestamp=None):

        if self.__domain == RFStormDomains.RFSTORM_RAW_ESB:
            self._send_whad_esb_pdu(pdu, address, timestamp)
        elif self.__domain == RFStormDomains.RFSTORM_UNIFYING:
            self._send_whad_unifying_pdu(pdu, address, timestamp)
        elif self.__domain == RFStormDomains.RFSTORM_PHY:
            self._send_whad_phy_pdu(address + pdu, timestamp)

    def _send_whad_esb_pdu(self, pdu, address, timestamp=None):
        msg = self.hub.esb.create_pdu_received(
            self.__channel,
            pdu,
            address=ESBAddress(address)
        )

        # Set timestamp if provided
        if timestamp is not None:
            msg.timestamp = timestamp

        # Send message
        self._send_whad_message(msg)

    def _send_whad_unifying_pdu(self, pdu, address, timestamp=None):
        msg = self.hub.unifying.create_pdu_received(
            self.__channel,
            pdu,
            address=ESBAddress(address)
        )

        # Set timestamp if provided
        if timestamp is not None:
            msg.timestamp = timestamp


        # Send message
        self._send_whad_message(msg)

    def _send_whad_phy_pdu(self, packet, timestamp=None):
        msg = self.hub.phy.create_packet_received(
            (self.__channel + 2400) * 1000000,
            packet
        )

        # Set timestamp if provided
        if timestamp is not None:
            msg.timestamp = timestamp

        # Send message
        self._send_whad_message(msg)

    def _on_whad_phy_sync_word(self, message):
        if len(message.sync_word) > 0 and len(message.sync_word) < 5:
            self.__phy_sync = message.sync_word
            self._send_whad_command_result(CommandResult.SUCCESS)
        else:
            self._send_whad_command_result(CommandResult.PARAMETER_ERROR)

    def _on_whad_phy_get_supported_freq(self, message):
        ranges = (self.__supported_frequency_range[0], self.__supported_frequency_range[1])

        # Create a SupportedFreqRanges response message
        msg = self.hub.phy.create_supported_freq_ranges([ranges])

        # Send message
        self._send_whad_message(msg)

    def _on_whad_phy_set_freq(self, message):
        range_start, range_end = self.__supported_frequency_range[0:2]

        if message.frequency >= range_start and message.frequency <= range_end:
            self.__channel = int(message.frequency / 1000000) - 2400
            self._send_whad_command_result(CommandResult.SUCCESS)
        else:
            self._send_whad_command_result(CommandResult.PARAMETER_ERROR)


    def _on_whad_phy_datarate(self, message):
        rates = {
            250000 : RFStormDataRate.RF_250KBPS,
            1000000 : RFStormDataRate.RF_1MBPS,
            2000000 : RFStormDataRate.RF_2MBPS,
        }
        if message.rate in rates:
            self.__phy_rate = rates[message.rate]
            self._send_whad_command_result(CommandResult.SUCCESS)
        else:
            self._send_whad_command_result(CommandResult.PARAMETER_ERROR)


    def _on_whad_phy_packet_size(self, message):
        if message.packet_size >= 5 and message.packet_size <= 31:
            self.__phy_size = message.packet_size
            self._send_whad_command_result(CommandResult.SUCCESS)
        else:
            self._send_whad_command_result(CommandResult.PARAMETER_ERROR)

    def _on_whad_phy_endianness(self, message):
        if message.endianness in (Endianness.BIG, Endianness.LITTLE):
            self.__phy_endianness = message.endianness
            self._send_whad_command_result(CommandResult.SUCCESS)
        else:
            self._send_whad_command_result(CommandResult.PARAMETER_ERROR)

    def _on_whad_phy_mod_gfsk(self, message):
        self._send_whad_command_result(CommandResult.SUCCESS)


    def _on_whad_phy_send(self, message):
        self._on_whad_send(message)

    def _on_whad_send(self, message):
        if self.__domain == RFStormDomains.RFSTORM_PHY:
            if self.__phy_endianness == Endianness.LITTLE:
                sync = bytes([swap_bits(i) for i in self.__phy_sync])[::-1]
                data =  bytes([swap_bits(i) for i in message.packet])
            else:
                sync = self.__phy_sync
                data = message.packet
            success = self._rfstorm_transmit_payload_generic(sync + data, address=b"")
            if success:
                self._send_whad_command_result(CommandResult.SUCCESS)
            else:
                self._send_whad_command_result(CommandResult.PARAMETER_ERROR)

        else:
            #channel = message.channel if message.channel != 0xFF else self.__channel
            pdu = message.pdu
            retransmission_count = message.retr_count
            if self.__acking:
                self.__ack_payload = pdu
            else:
                ack = self._rfstorm_transmit_payload(pdu, retransmits=retransmission_count)
                self._send_whad_command_result(CommandResult.SUCCESS)

                if self.__check_ack:
                    if ack:
                        self._send_whad_pdu(b"", address=self.__address)
                        self._send_whad_command_result(CommandResult.SUCCESS)
                    else:
                        self._send_whad_command_result(CommandResult.SUCCESS)

            self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_esb_send(self, message):
        self.__domain = RFStormDomains.RFSTORM_RAW_ESB
        self._on_whad_send(message)

    def _on_whad_unifying_send(self, message):
        self.__domain = RFStormDomains.RFSTORM_UNIFYING
        self._on_whad_send(message)

    def _on_whad_set_node_addr(self, message):
        self.__address = message.address
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_esb_set_node_addr(self, message):
        self.__domain = RFStormDomains.RFSTORM_RAW_ESB
        self._on_whad_set_node_addr(message)

    def _on_whad_unifying_set_node_addr(self, message):
        self.__domain = RFStormDomains.RFSTORM_UNIFYING
        self._on_whad_set_node_addr(message)

    def _on_whad_ptx(self, message):
        self.__internal_state = RFStormInternalStates.SNIFFING
        self.__acking = False
        self.__check_ack = True
        self.__ptx = True
        self.__channel = message.channel
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_esb_ptx(self, message):
        self.__domain = RFStormDomains.RFSTORM_RAW_ESB
        self._on_whad_ptx(message)

    def _on_whad_unifying_mouse(self, message):
        self.__domain = RFStormDomains.RFSTORM_UNIFYING
        self._on_whad_ptx(message)

    def _on_whad_unifying_keyboard(self, message):
        self.__domain = RFStormDomains.RFSTORM_UNIFYING
        self._on_whad_ptx(message)

    def _on_whad_prx(self, message):
        self.__internal_state = RFStormInternalStates.SNIFFING
        self.__acking = True
        self.__check_ack = False
        self.__ptx = False
        self.__channel = message.channel
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_esb_prx(self, message):
        self.__domain = RFStormDomains.RFSTORM_RAW_ESB
        self._on_whad_prx(message)

    def _on_whad_unifying_dongle(self, message):
        self.__domain = RFStormDomains.RFSTORM_UNIFYING
        self._on_whad_prx(message)

    def _on_whad_sniff(self, message):
        if self.__domain == RFStormDomains.RFSTORM_PHY:
            self.__internal_state = RFStormInternalStates.SNIFFING

            self._send_whad_command_result(CommandResult.SUCCESS)
        else:
            channel = message.channel
            show_acknowledgements = message.show_acks
            address = message.address
            self.__ptx = False
            self.__channel = channel

            self.__acking = False
            if address == b"\xFF\xFF\xFF\xFF\xFF":
                self.__internal_state = RFStormInternalStates.PROMISCUOUS_SNIFFING
            else:
                self.__internal_state = RFStormInternalStates.SNIFFING
                self.__address = address
            self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_esb_sniff(self, message):
        self.__domain = RFStormDomains.RFSTORM_RAW_ESB
        self._on_whad_sniff(message)

    def _on_whad_unifying_sniff(self, message):
        self.__domain = RFStormDomains.RFSTORM_UNIFYING
        self._on_whad_sniff(message)

    def _on_whad_phy_sniff(self, message):
        self.__domain = RFStormDomains.RFSTORM_PHY
        self._on_whad_sniff(message)

    def _on_whad_stop(self, message):
        self.__opened_stream = False
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_esb_stop(self, message):
        self.__domain = RFStormDomains.RFSTORM_RAW_ESB
        self._on_whad_stop(message)

    def _on_whad_unifying_stop(self, message):
        self.__domain = RFStormDomains.RFSTORM_UNIFYING
        self._on_whad_stop(message)

    def _on_whad_phy_stop(self, message):
        self.__domain = RFStormDomains.RFSTORM_PHY
        self._on_whad_stop(message)

    def _on_whad_start(self, message):
        self._rfstorm_enable_lna()
        if self.__domain == RFStormDomains.RFSTORM_PHY:
            if self.__phy_endianness == Endianness.LITTLE:
                sync = bytes([swap_bits(i) for i in self.__phy_sync])[::-1]
            else:
                sync = self.__phy_sync

            # Set channel
            success_chan = self._rfstorm_set_channel(self.__channel)

            # Enable promiscuous mode
            success = self._rfstorm_generic_promiscuous_mode(prefix=sync, rate=self.__phy_rate,
                                                             payload_length=self.__phy_size)
            if success and success_chan:
                self.__opened_stream = True
                self._send_whad_command_result(CommandResult.SUCCESS)
            else:
                self._send_whad_command_result(CommandResult.ERROR)

        else:
            if self.__channel == 0xFF:
                self.__scanning = True
                self.__channel = 0
            else:
                self.__scanning = False
            success = self._rfstorm_set_channel(self.__channel)
            if self.__internal_state == RFStormInternalStates.SNIFFING:
                success = success and self._rfstorm_sniffer_mode(self.__address[::-1])
            elif self.__internal_state == RFStormInternalStates.PROMISCUOUS_SNIFFING:
                success = success and self._rfstorm_promiscuous_mode()

            if success:
                self.__opened_stream = True
                self._send_whad_command_result(CommandResult.SUCCESS)
            else:
                self._send_whad_command_result(CommandResult.ERROR)

    def _on_whad_esb_start(self, message):
        self.__domain = RFStormDomains.RFSTORM_RAW_ESB
        self._on_whad_start(message)

    def _on_whad_unifying_start(self, message):
        self.__domain = RFStormDomains.RFSTORM_UNIFYING
        self._on_whad_start(message)

    def _on_whad_phy_start(self, message):
        self.__domain = RFStormDomains.RFSTORM_PHY
        self._on_whad_start(message)



"""
This module provides some constants used by WHAD to communicate with the APIMote.
"""
from enum import IntEnum
from dataclasses import dataclass


@dataclass
class RegisterMask:
    """Register mask class
    """
    mask : int = 0
    offset : int = 0

class APIMoteId(IntEnum):
    """USB product and vendor IDs.
    """
    APIMOTE_ID_VENDOR    = 0x0403
    APIMOTE_ID_PRODUCT   = 0x6015

class APIMoteInternalStates(IntEnum):
    """APIMote internal states.
    """
    NONE                    = 0
    SNIFFING                = 1
    TRANSMITTING            = 2


class APIMoteRegisters(IntEnum):
    """APIMote registers offsets.
    """
    # Status registers: initiate the change of an internal state or mode
    SNOP            = 0x00        # No operation
    SXOSCON         = 0x01        # Crystal oscillator
    STXCAL          = 0x02        # Enable/calibrate frequency Synthesizer
    SRXON           = 0x03        # Enable RX
    STXON           = 0x04        # Enable TX
    STXONCCA        = 0x05        # Enable TX if CCA indicates a clear channel
    SRFOFF          = 0x06        # Disable RX/TX and frequency synthetiser
    SRXOSCOFF       = 0x07        # Turn off RF and crystal oscillator
    SFLUSHRX        = 0x08        # Flush RX fifo and reset demodulator
    SFLUSHTX        = 0x09        # Flush TX fifo
    SACK            = 0x0A        # Send Acknowledgment, with pending field cleared
    SACKPEND        = 0x0B        # Send Acknowledgment, with pending field set
    SRXDEC          = 0x0C        # Start RX fifo decryption / authentication
    STXENC          = 0x0D        # Start TX fifo encryption / authentication
    SAES            = 0x0E        # AES encryption strobe

    # Configuration registers
    MAIN            = 0x10        # Main Control Register
    MDMCTRL0        = 0x11        # Modem Control Register 0
    MDMCTRL1        = 0x12        # Modem Control Register 1
    RSSI            = 0x13        # RSSI and CCA Status and Control
    SYNCWORD        = 0x14        # Synchronisation word control register
    TXCTRL          = 0x15        # Transmit Control Register
    RXCTRL0         = 0x16        # Receive Control Register 0
    RXCTRL1         = 0x17        # Receive Control Register 1
    FSCTRL          = 0x18        # Frequency Synthesizer Control and Status Register
    SECCTRL0        = 0x19        # Security Control Register 0
    SECCTRL1        = 0x1A        # Security Control Register 1
    BATTMON         = 0x1B        # Battery Monitor Control and Status Register
    IOCFG0          = 0x1C        # Input / Output Control Register 0
    IOCFG1          = 0x1D        # Input / Output Control Register 1
    MANFIDL         = 0x1E        # Manufacturer ID, Low 16 bits
    MANFIDH         = 0x1F        # Manufacturer ID, High 16 bits
    FSMTC           = 0x20        # Finite State Machine Time Constants
    MANAND          = 0x21        # Manual signal AND override register
    MANOR           = 0x22        # Manual signal OR override register
    AGCCTRL         = 0x23        # AGC Control Register
    AGCTST0         = 0x24        # AGC Test Register 0
    AGCTST1         = 0x25        # AGC Test Register 1
    AGCTST2         = 0x26        # AGC Test Register 2
    FSTST0          = 0x27        # Frequency Synthesizer Test Register 0
    FSTST1          = 0x28        # Frequency Synthesizer Test Register 1
    FSTST2          = 0x29        # Frequency Synthesizer Test Register 2
    FSTST3          = 0x2A        # Frequency Synthesizer Test Register 3
    RXBPFTST        = 0x2B        # Receiver Bandpass Filter Test Register
    FSMSTATE        = 0x2C        # Finite State Machine State Status Register
    ADCTST          = 0x2D        # ADC Test Register
    DACTST          = 0x2E        # DAC Test Register
    TOPTST          = 0x2F        # Top Level Test Register
    RESERVED        = 0x30        # Reserved for future use control
    TXFIFO          = 0x3E        # Transmit FIFO Byte Register
    RXFIFO          = 0x3F        # Receiver FIFO Byte Register

class APIMoteRegistersMasks:
    """APIMote registers masks definition.
    """

    class MCSM0:
        """Mode control register 0
        """
        CLOSE_IN_RX = RegisterMask(mask=0b11, offset=0)
        FS_AUTOCAL = RegisterMask(mask=0b11, offset=4)

    class MCSM1:
        """Mode control register 1
        """
        TX_OFF_MODE = RegisterMask(mask=0b11, offset=0)
        RX_OFF_MODE = RegisterMask(mask=0b11, offset=2)
        CCA_MODE = RegisterMask(mask=0b11, offset=4)

    class MDMCTRL0:
        """Modem control register 0
        """
        PREAMBLE_LENGTH = RegisterMask(mask=0b1111, offset=0)
        AUTO_ACK = RegisterMask(mask=0b1, offset=4)
        AUTO_CRC = RegisterMask(mask=0b1, offset=5)
        CCA_MODE = RegisterMask(mask=0b11, offset=6)
        CCA_HYST = RegisterMask(mask=0b111, offset=8)
        ADR_DECODE = RegisterMask(mask=0b1, offset=11)
        PAN_COORDINATOR = RegisterMask(mask=0b1, offset=12)
        RESERVED_FRAME_MODE = RegisterMask(mask=0b1, offset=13)

    class MDMCTRL1:
        """Modem control register 1
        """
        RX_MODE = RegisterMask(mask=0b11, offset=0)
        TX_MODE = RegisterMask(mask=0b11, offset=2)
        MODULATION_MODE = RegisterMask(mask=0b1, offset=4)
        DEMOD_AVG_MODE = RegisterMask(mask=0b1, offset=5)
        CORR_THR = RegisterMask(mask=0b11111, offset=6)

    class IOCFG0:
        """Input/Output configuration control register 0
        """
        FIFOP_THR = RegisterMask(mask=0b1111111, offset=0)
        CCA_POLARITY = RegisterMask(mask=0b1, offset=7)
        SFD_POLARITY = RegisterMask(mask=0b1, offset=8)
        FIFOP_POLARITY = RegisterMask(mask=0b1, offset=9)
        FIFO_POLARITY = RegisterMask(mask=0b1, offset=10)
        BCN_ACCEPT = RegisterMask(mask=0b1, offset=11)

    class SECCTRL0:
        """Security Controller control register 0
        """
        SEC_MODE = RegisterMask(mask=0b11, offset=0)
        SEC_M = RegisterMask(mask=0b111, offset=2)
        SEC_RXKEYSEL = RegisterMask(mask=0b1, offset=5)
        SEC_TXKEYSEL = RegisterMask(mask=0b1, offset=6)
        SEC_SAKEYSEL = RegisterMask(mask=0b1, offset=7)
        SEC_CBC_HEAD = RegisterMask(mask=0b1, offset=8)
        RXFIFO_PROTECTION = RegisterMask(mask=0b1, offset=9)

    class FSCTRL:
        """Frequency Selection Control register
        """
        FREQ = RegisterMask(mask=0b1111111111, offset=0)
        LOCK_STATUS = RegisterMask(mask=0b1, offset=10)
        LOCK_LENGTH = RegisterMask(mask=0b1, offset=11)
        CAL_RUNNING = RegisterMask(mask=0b1, offset=12)
        CAL_DONE = RegisterMask(mask=0b1, offset=13)
        LOCK_THR = RegisterMask(mask=0b11, offset=14)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: whad/protocol/esb/esb.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1bwhad/protocol/esb/esb.proto\x12\x03\x65sb\"$\n\x11SetNodeAddressCmd\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x0c\"K\n\x08SniffCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\x12\x0f\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0c\x12\x1d\n\x15show_acknowledgements\x18\x03 \x01(\x08\"\x19\n\x06JamCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\"E\n\x07SendCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\x12\x1c\n\x14retransmission_count\x18\x02 \x01(\r\x12\x0b\n\x03pdu\x18\x03 \x01(\x0c\"H\n\nSendRawCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\x12\x1c\n\x14retransmission_count\x18\x02 \x01(\r\x12\x0b\n\x03pdu\x18\x03 \x01(\x0c\")\n\x16PrimaryReceiverModeCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\",\n\x19PrimaryTransmitterModeCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\"\n\n\x08StartCmd\"\t\n\x07StopCmd\"\x1b\n\x06Jammed\x12\x11\n\ttimestamp\x18\x01 \x01(\x04\"\xbe\x01\n\x0eRawPduReceived\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\x12\x11\n\x04rssi\x18\x02 \x01(\x05H\x00\x88\x01\x01\x12\x16\n\ttimestamp\x18\x03 \x01(\x04H\x01\x88\x01\x01\x12\x19\n\x0c\x63rc_validity\x18\x04 \x01(\x08H\x02\x88\x01\x01\x12\x14\n\x07\x61\x64\x64ress\x18\x05 \x01(\x0cH\x03\x88\x01\x01\x12\x0b\n\x03pdu\x18\x06 \x01(\x0c\x42\x07\n\x05_rssiB\x0c\n\n_timestampB\x0f\n\r_crc_validityB\n\n\x08_address\"\xbb\x01\n\x0bPduReceived\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\x12\x11\n\x04rssi\x18\x02 \x01(\x05H\x00\x88\x01\x01\x12\x16\n\ttimestamp\x18\x03 \x01(\x04H\x01\x88\x01\x01\x12\x19\n\x0c\x63rc_validity\x18\x04 \x01(\x08H\x02\x88\x01\x01\x12\x14\n\x07\x61\x64\x64ress\x18\x05 \x01(\x0cH\x03\x88\x01\x01\x12\x0b\n\x03pdu\x18\x06 \x01(\x0c\x42\x07\n\x05_rssiB\x0c\n\n_timestampB\x0f\n\r_crc_validityB\n\n\x08_address\"\xc1\x03\n\x07Message\x12/\n\rset_node_addr\x18\x01 \x01(\x0b\x32\x16.esb.SetNodeAddressCmdH\x00\x12\x1e\n\x05sniff\x18\x02 \x01(\x0b\x32\r.esb.SniffCmdH\x00\x12\x1a\n\x03jam\x18\x03 \x01(\x0b\x32\x0b.esb.JamCmdH\x00\x12\x1c\n\x04send\x18\x04 \x01(\x0b\x32\x0c.esb.SendCmdH\x00\x12#\n\x08send_raw\x18\x05 \x01(\x0b\x32\x0f.esb.SendRawCmdH\x00\x12*\n\x03prx\x18\x06 \x01(\x0b\x32\x1b.esb.PrimaryReceiverModeCmdH\x00\x12-\n\x03ptx\x18\x07 \x01(\x0b\x32\x1e.esb.PrimaryTransmitterModeCmdH\x00\x12\x1e\n\x05start\x18\x08 \x01(\x0b\x32\r.esb.StartCmdH\x00\x12\x1c\n\x04stop\x18\t \x01(\x0b\x32\x0c.esb.StopCmdH\x00\x12\x1d\n\x06jammed\x18\n \x01(\x0b\x32\x0b.esb.JammedH\x00\x12&\n\x07raw_pdu\x18\x0b \x01(\x0b\x32\x13.esb.RawPduReceivedH\x00\x12\x1f\n\x03pdu\x18\x0c \x01(\x0b\x32\x10.esb.PduReceivedH\x00\x42\x05\n\x03msg*\x95\x01\n\nESBCommand\x12\x12\n\x0eSetNodeAddress\x10\x00\x12\t\n\x05Sniff\x10\x01\x12\x07\n\x03Jam\x10\x02\x12\x08\n\x04Send\x10\x03\x12\x0b\n\x07SendRaw\x10\x04\x12\x17\n\x13PrimaryReceiverMode\x10\x05\x12\x1a\n\x16PrimaryTransmitterMode\x10\x06\x12\t\n\x05Start\x10\x07\x12\x08\n\x04Stop\x10\x08\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'whad.protocol.esb.esb_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _ESBCOMMAND._serialized_start=1300
  _ESBCOMMAND._serialized_end=1449
  _SETNODEADDRESSCMD._serialized_start=36
  _SETNODEADDRESSCMD._serialized_end=72
  _SNIFFCMD._serialized_start=74
  _SNIFFCMD._serialized_end=149
  _JAMCMD._serialized_start=151
  _JAMCMD._serialized_end=176
  _SENDCMD._serialized_start=178
  _SENDCMD._serialized_end=247
  _SENDRAWCMD._serialized_start=249
  _SENDRAWCMD._serialized_end=321
  _PRIMARYRECEIVERMODECMD._serialized_start=323
  _PRIMARYRECEIVERMODECMD._serialized_end=364
  _PRIMARYTRANSMITTERMODECMD._serialized_start=366
  _PRIMARYTRANSMITTERMODECMD._serialized_end=410
  _STARTCMD._serialized_start=412
  _STARTCMD._serialized_end=422
  _STOPCMD._serialized_start=424
  _STOPCMD._serialized_end=433
  _JAMMED._serialized_start=435
  _JAMMED._serialized_end=462
  _RAWPDURECEIVED._serialized_start=465
  _RAWPDURECEIVED._serialized_end=655
  _PDURECEIVED._serialized_start=658
  _PDURECEIVED._serialized_end=845
  _MESSAGE._serialized_start=848
  _MESSAGE._serialized_end=1297
# @@protoc_insertion_point(module_scope)



from scapy.packet import Packet, bind_layers, split_layers
from scapy.fields import ByteEnumField, ShortEnumField, IntField, XShortField, \
    UUIDField, StrFixedLenField, ConditionalField, PacketField, XLongField, XIntField, \
    BitEnumField, BitField
from scapy.layers.bluetooth import EIR_Element, EIR_Hdr, EIR_Raw




class BTMesh_Unprovisioned_Device_Beacon(Packet):
    name = "Bluetooth Mesh Unprovisioned Device Beacon"
    fields_desc = [
        UUIDField("device_uuid", None, uuid_fmt=UUIDField.FORMAT_BE),
        ShortEnumField("oob_information", None, {
            0:"Other",
            1:"Electronic / URI",
            2:"2D machine-readable code",
            3:"Bar code",
            4:"Near Field Communication (NFC)",
            5:"Number",
            6:"String",
            7:"Support for certificate-based provisioning",
            8:"Support for provisioning records",
            9:"Reserved for Future Use",
            10:"Reserved for Future Use",
            11:"On box",
            12:"Inside box",
            13:"On piece of paper",
            14:"Inside manual",
            15:"On device",
        } ),
        IntField("uri_hash", None)
    ]
class BTMesh_Secure_Network_Beacon(Packet):
    name = "Bluetooth Mesh Secure Network Beacon"
    fields_desc = [
        BitEnumField("key_refresh_flag", 0,1,[False, True]),
        BitEnumField("iv_update_flag", 0,1,["normal_operation", "iv_update_in_progress"]),
        BitField("unused", 0,6),
        XLongField("network_id", None),
        XIntField("iv_index", None),
        XLongField("authentication_value", None)

    ]

class BTMesh_Private_Beacon(Packet):
    name = "Bluetooth Mesh Private Beacon"
    fields_desc = [
        StrFixedLenField('random', None, length=13),
        StrFixedLenField('obfuscated_private_beacon_data', None, length=5),
        XLongField("authentication_tag", None)

    ]
class EIR_BTMesh_Beacon(EIR_Element):
    name = "Bluetooth Mesh Beacon"
    fields_desc = [
        ByteEnumField("mesh_beacon_type", None, {
            0x00 : "unprovisioned_device_beacon",
            0x01 : "secure_network_beacon",
            0x02 : "mesh_private_beacon"
        }),
        ConditionalField(
            PacketField("unprovisioned_device_beacon_data", None, pkt_cls=BTMesh_Unprovisioned_Device_Beacon)
        , lambda pkt:pkt.mesh_beacon_type == 0),
        ConditionalField(
            PacketField("secure_beacon_data", None, pkt_cls=BTMesh_Secure_Network_Beacon)
        , lambda pkt:pkt.mesh_beacon_type == 1),
        ConditionalField(
            PacketField("private_beacon_data", None, pkt_cls=BTMesh_Private_Beacon)
        , lambda pkt:pkt.mesh_beacon_type == 2)
    ]

split_layers(EIR_Hdr, EIR_Raw)
bind_layers(EIR_Hdr, EIR_BTMesh_Beacon, type=0x2b)
bind_layers(EIR_Hdr, EIR_Raw)



"""
Pythonic Bluetooth LE stack
"""
from .llm import LinkLayer
from .constants import BtVersion

from whad.common.stack import Layer, alias, source

import logging
logger = logging.getLogger(__name__)

@alias('phy')
class BleStack(Layer):

    def __init__(self, connector, bt_version=BtVersion(4, 0), manufacturer=0x0002, sub_version=0x0100, options={}):
        super().__init__(options=options)

        # Save connector (used as PHY layer)
        self.__connector = connector

        # Store BT supported version, manufacturer and sub version
        self.__version = bt_version
        self.__manufacturer = manufacturer
        self.__sub_version = sub_version

    @property
    def bt_version(self):
        return self.__version

    @property
    def manufacturer_id(self):
        return self.__manufacturer

    @property
    def bt_sub_version(self):
        return self.__sub_version

    def on_connection(self, conn_handle, local_peer_addr, remote_peer_addr):
        '''BLE connection callback.
        '''
        # Call the LL layer object's `on_connect()` method to give it all the
        # required information
        connection = self.get_layer('ll').on_connect(
            conn_handle,
            local_peer_addr,
            remote_peer_addr
        )

        # Tell the WHAD connector we have a new connection.
        self.__connector.on_new_connection(connection)

    def on_disconnection(self, conn_handle, reason):
        '''BLE disconnection callback.
        '''
        # Notify link layer that a connection has been terminated.
        self.get_layer('ll').on_disconnect(conn_handle)

    def on_ctl_pdu(self, conn_handle, pdu):
        '''Control PDU callback.

        This callback handles a control PDU received from a connection
        identified by its connection handle `conn_handle`.
        '''
        logger.debug("received a control PDU (%d bytes) for connection handle %d", len(pdu), conn_handle)
        # Notify link layer we received a control PDU for a given `conn_handle`.
        self.send('ll', pdu, tag='control', conn_handle=conn_handle)


    def on_data_pdu(self, conn_handle, pdu):
        '''Data PDU callback.

        This callback hanles a data PDU received from a connection identitied
        by its connection handle `conn_handle`.
        '''
        logger.debug("received a data PDU (%d bytes) for conn_handle %d", len(pdu), conn_handle)
        # Notify link layer we received a data PDU for a given `conn_handle`.
        self.send('ll', pdu, tag='data', conn_handle=conn_handle)

    @source('ll', 'data')
    def send_data(self, data, conn_handle=None, encrypt=None):
        '''Send data to the underlying WHAD connector.
        '''
        logger.debug("sending a data PDU (%d bytes) to conn_handle %d", len(data), conn_handle)
        return self.__connector.send_data_pdu(data, conn_handle=conn_handle, encrypt=encrypt)

    @source('ll', 'control')
    def send_control(self, pdu, conn_handle=None, encrypt=None):
        '''Send control PDU to the underlying WHAD connector.
        '''
        logger.debug("sending a control PDU (%d bytes) to conn_handle %d", len(pdu), conn_handle)
        self.__connector.send_ctrl_pdu(pdu, conn_handle, encrypt=encrypt)

    @source('ll', tag='ATT_MTU')
    def on_mtu_changed(self, mtu, conn_handle=None):
        """Notify connector that the MTU has changed.
        """
        self.__connector.on_mtu_changed(conn_handle, mtu)

    def set_encryption(self, conn_handle=None, enabled=True,ll_key=None, ll_iv=None, key=None, rand=None, ediv=None):
        '''Enable or disable encryption using underlying WHAD connector.
        '''
        logger.debug("%s encryption (key=%s, iv=%s)", "enabling" if enabled else "disabling", ll_key.hex(), ll_iv.hex())
        return self.__connector.set_encryption(
            conn_handle=conn_handle,
            enabled=enabled,
            ll_key=ll_key,
            ll_iv=ll_iv,
            key=key,
            rand=rand,
            ediv=ediv
        )

BleStack.add(LinkLayer)



"""WHAD Protocol Generic message abstraction layer.
"""
from whad.protocol.whad_pb2 import Message
from whad.hub.message import HubMessage, pb_bind, Registry
from whad.hub import ProtocolHub

@pb_bind(ProtocolHub, name='generic', version=1)
class Generic(Registry):
    """WHAD Generic message parser/factory.
    """

    NAME = 'generic'
    VERSIONS = {}

    def __init__(self, version: int):
        self.proto_version = version

    @staticmethod
    def parse(proto_version: int, message) -> HubMessage:
        """Parses a WHAD generic message as seen by protobuf
        """
        message_type = message.generic.WhichOneof('msg')
        message_clazz = Generic.bound(message_type, proto_version)
        return message_clazz.parse(proto_version, message)

    def create_command_result(self, result_code: int) -> HubMessage:
        """Create a generic command result.
        """
        # Map our result code to our dedicated methods
        result_types = {
            CommandResult.SUCCESS: self.create_success,
            CommandResult.ERROR: self.create_error,
            CommandResult.PARAMETER_ERROR: self.create_param_error,
            CommandResult.BUSY: self.create_busy,
            CommandResult.DISCONNECTED: self.create_disconnected,
            CommandResult.WRONG_MODE: self.create_wrong_mode,
        }

        # Dispatch to our specific factory methods for command result.
        if result_code in result_types:
            return result_types[result_code]()
        else:
            raise ValueError()

    def create_error(self) -> HubMessage:
        """Create a generic error message.
        """
        return Generic.bound('cmd_result_error', self.proto_version)()

    def create_success(self) -> HubMessage:
        """Create a generic success result.
        """
        return Generic.bound('cmd_result_success', self.proto_version)()

    def create_param_error(self) -> HubMessage:
        """Create a parameter error message.
        """
        return Generic.bound('cmd_result_param_error', self.proto_version)()

    def create_disconnected(self) -> HubMessage:
        """Create a disconnected error message.
        """
        return Generic.bound('cmd_result_disconnected', self.proto_version)()

    def create_wrong_mode(self) -> HubMessage:
        """Create a wrong mode error message.
        """
        return Generic.bound('cmd_result_wrong_mode', self.proto_version)()

    def create_unsupported_domain(self) -> HubMessage:
        """Create an unsupported domain error message.
        """
        return Generic.bound('cmd_result_unsupported_domain', self.proto_version)()

    def create_busy(self) -> HubMessage:
        """Create a busy error message.
        """
        return Generic.bound('cmd_result_busy', self.proto_version)()

    def create_debug(self, level: int, message: bytes) -> HubMessage:
        """Create a debug message.
        """
        return Generic.bound('debug', self.proto_version)(
            debug_level=level, debug_msg=message
        )

    def create_verbose(self, message: bytes) -> HubMessage:
        """Create a verbose message.
        """
        return Generic.bound('verbose', self.proto_version)(
            data=message
        )

    def create_progress(self, value: int) -> HubMessage:
        """Create a progress message.
        """
        return Generic.bound('progress', self.proto_version)(
            value=value
        )

from .cmdresult import CommandResult
from .debug import Debug
from .verbose import Verbose
from .progress import Progress

__all__ = [
    'Generic',
    'CommandResult',
    'Debug',
    'Verbose',
    'Progress'
]



from whad.rf4ce.stack.apl.profile import APLProfile

class ZRCProfile(APLProfile):
    """
    APL service implementing the Zigbee Remote Control (ZRC) Profile.
    """

    def __init__(self):
        super().__init__(name="zrc", profile_id=0x01)


    def on_data(self, npdu, pairing_reference, vendor_id, link_quality, rx_flags):
        """
        Callback processing incoming data for the profile.
        """
        pass



"""LoRaWAN single-channel Gateway implementation.
"""
import logging

from whad.device import WhadDevice
from whad.lorawan.stack import LWGatewayStack
from whad.scapy.layers.lorawan import PHYPayload
from whad.lorawan.channel import ChannelPlan
from whad.lorawan.connector import LoRaWAN
from whad.lorawan.app import LWApplication

logger = logging.getLogger(__name__)

class LWGateway(LoRaWAN):
    """LoRaWAN single-channel gateway implementation.

    This emulated single-channel LoRaWAN gateway handles a single
    application.
    """

    def __init__(self, device : WhadDevice = None, channel_plan : ChannelPlan = None, app:LWApplication=None, stack=LWGatewayStack):
        """Initialize a LoRaWAN gateway.

        :param device: WHAD compatible device to use for this gateway
        :type device: WhadDevice
        :param channel_plan: LoRaWAN channel plan to use
        :type channel_plan: ChannelPlan
        :param app: LoRaWAN application to bind to this gateway
        :type app: LWApplication
        :param stack: LoRaWAN gateway stack class to use
        :type stack: class
        """
        super().__init__(device=device, channel_plan=channel_plan)

        self.__app = app
        self.__stack = stack(
            self,
            options={
                'appkey': app.key,
                'appeui': app.eui,
                'll': {

                }
            }
        )

        # Inject devices (if any) into our LoRaWAN link-layer
        linklayer = self.__stack.get_layer('ll')
        for device in self.__app.nodes():
            print(device)
            linklayer.add_provisioned_device(
                device.dev_addr,
                device.dev_eui,
                device.appskey,
                device.nwkskey,
                device.upcount,
                device.dncount
            )

        # Start RF
        self.start()

    def on_packet(self, packet : bytes):
        """Process incoming packet

        :param packet: Incoming packet
        :type packet: bytes
        """
        # Add packet to our packet queue
        pkt = PHYPayload(bytes(packet))
        pkt.metadata = packet.metadata
        self.__stack.on_frame(pkt)

    def is_device_allowed(self, dev_eui : str) -> bool:
        """Check if device is allowed to connect to our application.

        :param dev_eui: Device EUI
        :type dev_eui: str

        :return: True if device is allowed, False otherwise.
        :rtype: bool
        """
        # Forward to our application
        return self.__app.is_authorized(dev_eui)

    def on_device_joined(self, dev_eui : str, dev_addr : int, appskey : bytes, nwkskey : bytes):
        """Device joined callback

        This method processes an OTAA join event.

        :param dev_eui: Device EUI
        :type dev_eui: str
        :param dev_addr: Device network address
        :type dev_addr: int
        :param appskey: Device application session key
        :type appskey: bytes
        :param nwkskey: Device network encryption session key
        :type nwkske: bytes
        """
        logger.info(
            "[gateway] Device %s joined network with address 0x%08x",
            dev_eui,
            dev_addr,
        )

        self.__app.on_device_joined(
            dev_eui,
            dev_addr,
            appskey,
            nwkskey
        )
    
    def on_device_data(self, dev_eui : str, dev_addr : int, data : bytes, upcount : int) -> bytes:
        """Device incoming data callback

        :param dev_eui: Device EUI
        :type dev_eui: str
        :param dev_addr: Device network address
        :type dev_addr: int
        :param data: Data received by the device
        :type data: bytes
        :param upcount: uplink frame counter
        :type upcount: int

        :return: Data to be sent back to the device
        :rtype: bytes
        """
        logger.info("[gateway] Device %s sent data: %s", dev_eui, data.hex())
        return self.__app.on_device_data(
            dev_eui,
            dev_addr,
            data,
            upcount
        )

    def stop(self):
        """Stop the Gateway and its associated application.
        """
        self.__app.stop()
        super().stop()




"""
WHAD traffic analyzer base module.
"""
from typing import List, Any, Optional, Any, Type

from scapy.packet import Packet

class InvalidParameter(Exception):
    """Invalid parameter specified to a traffic analyzer."""
    def __init__(self, name: str, value: Any):
        """Initialization."""
        super().__init__()
        self.__name = name
        self.__value = value

    @property
    def parameter(self) -> str:
        """Invalid parameter name."""
        return self.__name

    @property
    def value(self) -> Any:
        """Parameter value"""
        return self.__value

    def __str__(self) -> str:
        """String representation of this exception."""
        return f"InvalidParameter(name=\"{self.__name}\", value=\"{self.__value}\")"

    def __repr__(self) -> str:
        """Representation of this exception."""
        return str(self)

class TrafficAnalyzer:
    """
    Traffic analyzer base class.

    This class must be inherited by specialized traffic analyzers
    in order to provide `wanalyze` with extracted and/or computed
    data.
    """

    # Traffic analyzer parameters (no parameters by default),
    # override this class property to define custom parameters.
    PARAMETERS = {}

    @classmethod
    def has_parameter(cls, parameter: str) -> bool:
        """ Check if a given parameter is expected, based on its name. """
        return parameter in cls.PARAMETERS

    @classmethod
    def get_default_parameters(cls) -> dict:
        """ Return the class default parameters. """
        default_params = {}
        for parameter, value in cls.PARAMETERS.items():
            default_params[parameter] = value
        return default_params

    def __init__(self, **kwargs):
        """ Initialize a traffic analyzer and set its associated
        configuration parameters, if provided.
        """
        # Filter parameters to only keep those we expect
        self.__parameters = self.get_default_parameters()
        for parameter,value in kwargs.items():
            if parameter in self.__parameters:
                self.__parameters[parameter] = value

        # Reset this traffic analyzer
        self.reset()

    def get_param(self, param: str) -> Any:
        """ Retrieve a provided parameter value. """
        if param in self.__parameters:
            return self.__parameters[param]

    def set_param(self, param: str, value: Any):
        """ Set traffic analyzer parameter value. """
        if param in self.__parameters:
            self.__parameters[param] = value

    def process_packet(self, packet: Packet):
        """Process a packet.
        """

    def mark_packet(self, packet):
        """Mark a specific packet.
        """
        self.__marked_packets.append(packet)

    def reset(self):
        """Reset traffic analyzer state.
        """
        self.__triggered = False
        self.__completed = False
        self.__marked_packets = []

    def trigger(self):
        """Trigger this traffic analyzer.
        """
        self.__triggered = True

    def complete(self):
        """Mark this analyzer as completed.

        Once a traffic analyzer is completed, its output
        can be queried.
        """
        self.__completed = True

    @property
    def marked_packets(self) -> List[Packet]:
        """Returns marked packets.
        """
        return self.__marked_packets

    @property
    def output(self) -> Optional[Any]:
        """Returns the traffic analyzer output.
        """
        return None

    @property
    def triggered(self) -> bool:
        """Determine if the traffic analyzer has been triggered.

        :rtype: bool
        :return: `True` if triggered, `False` otherwise.
        """
        return self.__triggered

    @property
    def completed(self) -> bool:
        """Determine if the traffic analyzer has completed is
        job.

        :rtype: bool
        :return: `True` if completed, `False` otherwise.
        """
        return self.__completed
